"""Actions."""

from construct import (Array, Byte, Const, CString, Flag, Float32l, If,
                       Int16ul, Int32sl, Int32ul, Padding, Peek, String,
                       Struct, this, Bytes, Embedded, IfThenElse)

from mgz.body.achievements import achievements
from mgz.enums import (DiplomacyStanceEnum, FormationEnum, GameActionModeEnum,
                       OrderTypeEnum, ReleaseTypeEnum, ResourceEnum,
                       ResourceLevelEnum, RevealMapEnum, StanceEnum,
                       AgeEnum, VictoryEnum)
from mgz.util import TimeSecAdapter, check_flags

# pylint: disable=invalid-name, bad-continuation

# Not all actions are defined, not all actions are complete.


interact = "interact"/Struct(
    "player_id"/Byte,
    Const(b"\x00\x00"),
    "target_id"/Int32ul,
    "selected"/Int32ul,
    "x"/Float32l,
    "y"/Float32l,
    "next"/Peek(Bytes(8)),
    "flags"/If(lambda ctx: check_flags(ctx.next), Bytes(8)),
    "unit_ids"/If(lambda ctx: ctx.selected < 0xff, Array(
        lambda ctx: ctx.selected, "unit_ids"/Int32ul
    ))
)

give_attribute = "give_attribute"/Struct(
    "player_id"/Byte,
    "target_id"/Byte,
    "attribute"/Byte,
    "amount"/Float32l
)

add_attribute = "add_attribute"/Struct(
    "player_id"/Byte,
    "attribute"/Byte,
    Padding(1),
    "amount"/Float32l
)

create = "create"/Struct(
    Padding(1),
    "unit_type"/Int16ul,
    "player_id"/Byte,
    Padding(1),
    "x"/Float32l,
    "y"/Float32l,
    "z"/Float32l
)

ai_interact = "ai_interact"/Struct(
    Padding(3),
    "target_id"/Int32ul,
    "selected"/Byte,
    Padding(3),
    "x"/Float32l,
    "y"/Float32l,
    If(lambda ctx: ctx.selected < 0xff, Array(
        lambda ctx: ctx.selected, "unit_ids"/Int32ul
    ))
)

move = "move"/Struct(
    "player_id"/Byte,
    Const(b"\x00\x00"),
    Padding(4),
    "selected"/Int32ul,
    "x"/Float32l,
    "y"/Float32l,
    If(lambda ctx: ctx.selected < 0xff, Struct(
        "next"/Peek(Bytes(8)),
        "flags"/If(lambda ctx: check_flags(ctx.next), Bytes(8))
    )),
    "unit_ids"/If(lambda ctx: ctx.selected < 0xff, Array(
        lambda ctx: ctx.selected, Int32ul
    ))
)

ai_move = "ai_move"/Struct(
    "selected"/Byte,
    "player_id"/Byte,
    "player_num"/Byte,
    Padding(4),
    Padding(4),
    "target_id"/Int32ul,
    Padding(1),
    Padding(3),
    "x"/Float32l,
    "y"/Float32l,
    Padding(4),
    Padding(4),
    Padding(4),
    If(lambda ctx: ctx.selected > 0x01, Array(
        lambda ctx: ctx.selected, "unit_ids"/Int32ul
    ))
)

resign = "resign"/Struct(
    "player_id"/Byte,
    "player_num"/Byte,
    "disconnected"/Flag
)

spec = "spec"/Struct(
    Padding(lambda ctx: ctx._._.length - 1)
)

queue = "queue"/Struct(
    Padding(3),
    "building_id"/Int32ul,
    "unit_type"/Int16ul,
    "number"/Int16ul,
)

multiqueue = "multiqueue"/Struct(
    Padding(3),
    "unit_type"/Int16ul,
    "num_buildings"/Byte,
    "queue_amount"/Byte,
    Array(lambda ctx: ctx.num_buildings, "building_ids"/Int32ul)
)

ai_queue = "ai_queue"/Struct(
    Padding(3),
    "building_id"/Int32ul,
    "player_id"/Int16ul,
    "unit_type"/Int16ul,
    Padding(4)
)

research = "research"/Struct(
    Padding(3),
    "building_id"/Int32ul,
    "player_id"/Int16ul,
    "next"/Peek(
        Struct(
            Padding(6),
            "check"/Int32sl
        )
    ),
    IfThenElse(lambda ctx: ctx.next.check == -1,
        Embedded(Struct(
            "selected"/Int16ul,
            "technology_type"/Int32ul,
            Array(lambda ctx: ctx.selected, "selected_ids"/Int32sl)
        )),
        Embedded(Struct(
            "technology_type"/Int16ul,
            Array(1, "selected_ids"/Int32sl)
        ))
    )
)

sell = "sell"/Struct(
    "player_id"/Byte,
    ResourceEnum("resource_type"/Byte),
    "amount"/Byte,
    Padding(4)
)

buy = "buy"/Struct(
    "player_id"/Byte,
    ResourceEnum("resource_type"/Byte),
    "amount"/Byte,
    Padding(4)
)

stop = "stop"/Struct(
    "selected"/Byte,
    Array(lambda ctx: ctx.selected, "object_ids"/Int32ul)
)

stance = "stance"/Struct(
    "selected"/Byte,
    StanceEnum("stance_type"/Byte),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

guard = "guard"/Struct(
    "selected"/Byte,
    Padding(2),
    "guarded_unit_id"/Int32ul,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

follow = "follow"/Struct(
    "selected"/Byte,
    Padding(2),
    "followed_unit_id"/Int32ul,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

formation = "formation"/Struct(
    "selected"/Byte,
    "player_id"/Int16ul,
    FormationEnum("formation_type"/Int32ul),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

save = "save"/Struct(
    "exited"/Flag,
    "player_id"/Byte,
    "filename"/CString(encoding='latin1'),
    Padding(lambda ctx: ctx._._.length - 23),
    "checksum"/Int32ul
)

chapter = "chapter"/Struct(
    "player_id"/Byte
)

build = "build"/Struct(
    "selected"/Byte,
    "player_id"/Int16ul,
    "x"/Float32l,
    "y"/Float32l,
    "building_type"/Int32ul,
    Padding(4),
    "sprite_id"/Int32ul,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

game = "game"/Struct(
    "mode"/GameActionModeEnum("mode_id"/Byte),
    "player_id"/Byte,
    Padding(1),
    "diplomacy"/If(this.mode == 'diplomacy', Struct(
        "target_player_id"/Byte,
        Padding(3),
        "stance_float"/Float32l,
        "stance"/DiplomacyStanceEnum("stance_id"/Byte),
    )),
    "speed"/If(this.mode == 'speed', Struct(
        Padding(4),
        "speed"/Float32l,
        Padding(1)
    )),
    "instant_build"/If(this.mode == 'instant_build', Struct(
        Padding(9)
    )),
    "quick_build"/If(this.mode == 'quick_build', Struct(
        "status"/Flag,
        Padding(8),
    )),
    "allied_victory"/If(this.mode == 'allied_victory', Struct(
        "player_id"/Byte,
        "status"/Flag,
        Padding(7)
    )),
    "cheat"/If(this.mode == 'cheat', Struct(
        "cheat_id"/Byte,
        Padding(8)
    )),
    "unk0"/If(this.mode == 'unk0', Struct(
        Padding(9)
    )),
    "spy"/If(this.mode == 'spy', Struct(
        Padding(9)
    )),
    "unk1"/If(this.mode == 'unk1', Struct(
        Padding(9)
    )),
    "farm_queue"/If(this.mode == 'farm_queue', Struct(
        "amount"/Byte, # this seems to be a bit inconsistent between versions, needs more research
        Padding(8)
    )),
    "farm_unqueue"/If(this.mode == 'farm_unqueue', Struct(
        "amount"/Byte, # this seems to be a bit inconsistent between versions, needs more research
        Padding(8)
    )),
    # toggle farm auto seed queue
    "farm_autoqueue"/If(this.mode == 'farm_autoqueue', Struct(
        Padding(9)
    )),

    "fishtrap_queue" / If(this.mode == 'fishtrap_queue', Struct(
        "amount" / Byte,
        Padding(8)
    )),
    "fishtrap_unqueue" / If(this.mode == 'fishtrap_unqueue', Struct(
        "amount" / Byte,
        Padding(8)
    )),

    # toggle fish trap auto place queue
    "fishtrap_autoqueue"/If(this.mode == 'fishtrap_autoqueue', Struct(
        Padding(9)
    )),

    # toggles the default stance when units are created. All players start on aggressive by default, if the player
    # (initially) has defensive enabled it is called right before the first unit is queued, and again every time
    # the player toggles it in the game options menu
    "default_stance" / If(this.mode == 'default_stance', Struct(
        Padding(9)
    )),
    Padding(3)
)

droprelic = "droprelic"/Struct(
    Const(b"\x00\x00\x00"),
    'unit_id'/Int32ul
)

wall = "wall"/Struct(
    "selected"/Byte,
    "player_id"/Byte,
    IfThenElse(lambda ctx: ctx._._.length - 16 - (ctx.selected * 4) == 8,
        Embedded(Struct(
            Padding(1),
            "start_x"/Int16ul,
            "start_y"/Int16ul,
            "end_x"/Int16ul,
            "end_y"/Int16ul,
            "building_id"/Int32ul,
            Padding(4),
            "flags"/Bytes(4)
        )),
        Embedded(Struct(
            "start_x"/Byte,
            "start_y"/Byte,
            "end_x"/Byte,
            "end_y"/Byte,
            Padding(1),
            "building_id"/Int32ul,
            Padding(4),
        ))
    ),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

delete = "delete"/Struct(
    Padding(3),
    "object_id"/Int32ul,
    "player_id"/Int32ul
)

attackground = "attackground"/Struct(
    "selected"/Byte,
    Padding(2),
    "x"/Float32l,
    "y"/Float32l,
    "next"/Peek(Bytes(4)),
    "flags"/If(lambda ctx: check_flags(ctx.next), Bytes(4)),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

tribute = "tribute"/Struct(
    "player_id"/Byte,
    "player_id_to"/Byte,
    ResourceEnum("resource_type"/Byte),
    "amount"/Float32l,
    "fee"/Float32l
)

repair = "repair"/Struct(
    "selected"/Byte,
    Padding(2),
    "repaired_id"/Int32ul,
    "next"/Peek(Bytes(4)),
    "flags"/If(lambda ctx: check_flags(ctx.next), Bytes(4)),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)


release = "release"/Struct(
    "selected"/Int16ul,
    Padding(1),
    "x"/Float32l, # -1 if none
    "y"/Float32l, # -1 if none
    ReleaseTypeEnum("release_type"/Byte),
    Padding(3),
    "release_id"/Int32ul,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

"""
unload = "unload"/Struct(
    "selected"/Int16ul,
    Padding(1),
    "x"/Float32l, # -1 if none
    "y"/Float32l, # -1 if none
    Padding(4),
    Padding(4), # 0xffffffff
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)
"""

togglegate = "togglegate"/Struct(
    Padding(3),
    "gate_id"/Int32ul
)

flare = "flare"/Struct(
    Padding(7),
    Array(9, "player_ids"/Byte),
    Padding(3),
    "x"/Float32l,
    "y"/Float32l,
    "player_id"/Byte,
    "player_number"/Byte,
    Padding(2)
)

order = "order"/Struct(
    "selected"/Byte,
    Padding(2),
    "building_id"/Int32sl, # -1 cancels production queue
    OrderTypeEnum("order_type"/Byte),
    "cancel_order"/Byte, # when cancelling production queue, this indicates which item in the queue is to be cancelled
    Padding(2),
    "x"/Float32l,
    "y"/Float32l,
    Padding(4), # const
    "next"/Peek(Bytes(4)),
    "flags"/If(lambda ctx: check_flags(ctx.next), Bytes(4)),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul),
)

gatherpoint = "gatherpoint"/Struct(
    "selected"/Byte,
    Padding(2),
    "target_id"/Int32ul,
    "target_type"/Int32ul,
    "x"/Float32l,
    "y"/Float32l,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)

townbell = "townbell"/Struct(
    Padding(3),
    "towncenter_id"/Int32ul,
    "active"/Int32ul
)

"""Patrol

10 X-coordinates followed by 10 Y-coordinates
First of each is popped off for consistency with other actions
"""
patrol = "patrol"/Struct(
    "selected"/Byte,
    "waypoints"/Int16ul,
    "x"/Float32l,
    Array(9, "x_more"/Float32l),
    "y"/Float32l,
    Array(9, "y_more"/Float32l),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul),
)

waypoint = "waypoint"/Struct(
    Padding(1),
    "selected"/Byte,
    "x"/Byte,
    "y"/Byte,
    "building_ids"/If(lambda ctx: ctx.selected != 255, Array(
        lambda ctx: ctx.selected, Int32ul
    ))
)

ai_waypoint = "ai_waypoint"/Struct(
    "selected"/Byte,
    "waypoint_count"/Byte,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul),
    Array(lambda ctx: ctx.waypoint_count, "x_more"/Byte),
    Array(lambda ctx: ctx.waypoint_count, "y_more"/Byte)
)

backtowork = "backtowork"/Struct(
    Padding(3),
    "towncenter_id"/Int32ul
)

ai_command = "ai_command"/Struct(
    Padding(lambda ctx: ctx._._.length - 1)
)

"""DE Queue

In DE queue and multi queue share the same command
"""
de_queue = "de_queue"/Struct(
    "player_id"/Byte,
    "building_type"/Int16ul,
    "selected"/Byte,
    Padding(1),
    "unit_type"/Int16ul,
    "queue_amount"/Byte,
    Padding(1),
    Array(lambda ctx: ctx.selected, "building_ids"/Int32ul)
)

"""DE Attack Move

It's almost the same as Patrol.

10 X-coordinates followed by 10 Y-coordinates
First of each is popped off for consistency with other actions
"""
de_attackmove = "de_attackmove"/Struct(
    "selected"/Byte,
    "waypoints"/Int16ul,
    "x"/Float32l,
    Array(9, "x_more"/Float32l),
    "y"/Float32l,
    Array(9, "y_more"/Float32l),
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul),
)

postgame = "achievements"/Struct(
    Padding(3),
    "scenario_filename"/String(32, padchar=b'\x00', trimdir='right', encoding='latin1'),
    "player_num"/Byte,
    "computer_num"/Byte,
    Padding(2),
    Peek("duration_int"/Int32ul),
    TimeSecAdapter("duration"/Int32ul),
    "cheats"/Flag,
    "complete"/Flag,
    Padding(2),
    "db_checksum"/Int32ul,
    "code_checksum"/Int32ul,
    "version"/Float32l,
    "map_size"/Byte,
    "map_id"/Byte,
    "population"/Int16ul,
    Peek("victory_type_id"/Byte),
    VictoryEnum("victory_type"/Byte),
    Peek("starting_age_id"/Byte),
    AgeEnum("starting_age"/Byte),
    Peek("starting_resources_id"/Byte),
    ResourceLevelEnum("starting_resources"/Byte),
    "all_techs"/Flag,
    "random_positions"/Flag,
    RevealMapEnum("reveal_map"/Byte),
    "is_deathmatch"/Flag,
    "is_regicide"/Flag,
    "starting_units"/Byte,
    "lock_teams"/Flag,
    "lock_speed"/Flag,
    Padding(1),
    Array(lambda ctx: ctx.player_num, achievements),
    Padding(4),
    Array(lambda ctx: (8 - ctx.player_num) * 63, Padding(4)),
)

de_autoscout = "de_autoscout"/Struct(
    "selected"/Byte,
    Array(lambda ctx: ctx.selected, "unit_ids"/Int32ul)
)



"""Fast(er) parsing for situations requiring speed."""
import io
import struct

from mgz.fast.enums import Operation, Action, Postgame, Age
from mgz.fast.actions import parse_action_71094
from mgz.util import check_flags, unpack


MAX_PLAYERS = 8
SYNC_LEN_PER_PLAYER = 11


def sync(data):
    """Handle synchronizations.

    There are two types of sync messages:
    1) Increment-only (millisecond time increment)
    2) Increment plus checksum

    Type 2 is further divided into pre and post-DE.

    DE extends the checksum with a variety of granular
    per player values. This structure was documented
    by yilinho in the following pull request:
    https://github.com/happyleavesaoc/aoc-mgz/pull/123

    To avoid requiring a version argument, this function
    does some readaheads and branches based on observed
    values.
    """
    increment, marker = struct.unpack('<II', data.read(8))
    checksum = None
    if marker:
        data.seek(-4, 1)
        return increment, None, {}
    checksum, is_de, = struct.unpack("<4xI4xI", data.read(16))
    if not is_de:
        data.read(8)
        return increment, checksum, {}
    data.seek(-16, 1)
    values = struct.unpack(f'<{MAX_PLAYERS * SYNC_LEN_PER_PLAYER}I', data.read(4 * MAX_PLAYERS * SYNC_LEN_PER_PLAYER))
    """
    These values are just guesses, so they might be incorrect.
    
    val  0: unknown. It seems this value is always 0.
    val  1: total resources(wood + food + gold + stone), probably round down
    val  2: unknown
    val  3: number of displayable objects, objects*2(with shadow?) + bodies + projectile, etc. exclude building foundations
    val  4: displayable object TTL in seconds + resource on villagers (there is sometime an object with 100 TTL and never die)
    val  5: unknown. It seems this value is always 0.
    val  6: number of objects (include foundation), when a unit is lost, this counter updates after its displayable objects(body) dies
    val  7: unknown
    val  8: player ID, from 1 to 8
    val  9: unknown
    val 10: unknown
    """
    checksum = sum(values)
    current_time, = struct.unpack('<I', data.read(4))  # duration from beginning in ms
    payload = {"current_time": current_time}
    for ptr in range(0, MAX_PLAYERS * SYNC_LEN_PER_PLAYER, SYNC_LEN_PER_PLAYER):
        if values[ptr + 1]:
            payload[values[ptr + 8]] = {
                'total_res': values[ptr + 1],
                'dp_obj_count': values[ptr + 3],
                'dp_obj_ttl': values[ptr + 4],
                'obj_count': values[ptr + 6],
            }
    return increment, checksum, payload


def viewlock(data):
    """Handle viewlocks."""
    x, y, _ = struct.unpack('<ffI', data.read(12))
    return x, y


def parse_action(action_type, data):
    """Parse player, objects, and coordinates from actions."""
    player_id, length = struct.unpack_from('<bh', data)
    if len(data) == length + 3:
        return parse_action_71094(action_type, player_id, data[3:])
    if action_type == Action.RESIGN:
        return dict(player_id=data[0])
    if action_type == Action.TRIBUTE:
        player_id, player_id_to, resource_id, amount, fee = struct.unpack_from('<bbbff', data)
        return dict(player_id=player_id, player_id_to=player_id_to, resource_id=resource_id, amount=amount, fee=fee)
    if action_type == Action.DE_TRIBUTE:
        player_id, player_id_to, unknown_byte, unknown_int, wood, food, gold, stone = struct.unpack_from('<bbbiffff', data)
        return dict(player_id=player_id,
                    player_id_to=player_id_to,
                    food=food,
                    wood=wood,
                    stone=stone,
                    gold=gold)
    if action_type == Action.MOVE:
        player_id, selected, x, y = struct.unpack_from('<b6xI2f', data)
        object_ids = []
        if selected != 255:
            offset = 0
            if check_flags(struct.unpack_from('<4b', data[19:])):
                offset = 4
            object_ids = struct.unpack_from('<' + str(selected) + 'I', data[19 + offset:])
        return dict(player_id=player_id, x=x, y=y, object_ids=list(object_ids))
    if action_type == Action.CREATE:
        player_id, x, y = struct.unpack_from('<3xhx2f', data)
        return dict(player_id=player_id, x=x, y=y)
    if action_type == Action.ORDER:
        player_id, target_id, selected, x, y = struct.unpack_from('<b2xIh2x2f', data)
        object_ids = []
        if selected != 255:
            offset = 0
            if check_flags(struct.unpack_from('<4b', data[19:])):
                offset = 4
            object_ids = struct.unpack_from('<' + str(selected) + 'I', data[19 + offset:])
        return dict(player_id=player_id, target_id=target_id, x=x, y=y, object_ids=list(object_ids))
    if action_type == Action.BUILD:
        player_id, x, y, building_id = struct.unpack_from('<xh2fI', data)
        return dict(player_id=player_id, x=x, y=y, building_id=building_id)
    if action_type == Action.STANCE:
        stance_id, *object_ids = struct.unpack_from('<xb' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), stance_id=stance_id)
    if action_type == Action.RESEARCH:
        object_id, player_id = struct.unpack_from('<3xIh', data)
        if len(data) >= 19:
            technology_id, = struct.unpack_from('<I', data[11:15])
        else:
            technology_id, = struct.unpack_from('<h', data[9:11])
        return dict(player_id=player_id, technology_id=technology_id, object_ids=[object_id])
    if action_type == Action.FORMATION:
        player_id, formation_id, *object_ids = struct.unpack_from('<xhI' + str(data[0]) + 'I', data)
        return dict(player_id=player_id, object_ids=list(object_ids), formation_id=formation_id)
    if action_type == Action.QUEUE:
        object_id, unit_id, amount = struct.unpack_from('<3xIhh', data)
        return dict(object_ids=[object_id], unit_id=unit_id, amount=amount)
    if action_type == Action.GATHER_POINT:
        target_id, x, y, *object_ids = struct.unpack_from('<3xi4x2f' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y, target_id=target_id)
    if action_type == Action.MULTIQUEUE:
        unit_id, amount, *object_ids = struct.unpack_from('<3xhxb' + str(data[5]) + 'I', data)
        return dict(object_ids=object_ids, unit_id=unit_id, amount=amount)
    if action_type == Action.PATROL:
        x, y, *object_ids = struct.unpack_from('<3xf36xf36x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.SPECIAL:
        target_id, order_id, x, y, *flags = struct.unpack_from('<3xib3x2f4x4b', data)
        offset = 0
        if check_flags(flags):
            offset = 4
        object_ids = struct.unpack_from('<' + str(data[0]) + 'I', data[23 + offset:])
        values = dict(object_ids=list(object_ids), order_id=order_id)
        if x > 0 and y > 0:
            values.update(dict(x=x, y=y))
        if target_id > 0:
            values.update(dict(target_id=target_id))
        return values
    if action_type == Action.BACK_TO_WORK:
        object_id, = struct.unpack_from('<3xI', data)
        return dict(object_ids=[object_id])
    if action_type == Action.UNGARRISON:
        selected, = struct.unpack_from('<h', data)
        x, y, *object_ids = struct.unpack_from('<3x2f8x' + str(selected) + 'I', data)
        if x > 0 and y > 0:
            return dict(object_ids=list(object_ids), x=x, y=y)
        return dict(object_ids=list(object_ids))
    if action_type == Action.BUY:
        player_id, resource_id, amount = struct.unpack_from('<bbb', data)
        return dict(player_id=player_id, resource_id=resource_id, amount=amount)
    if action_type == Action.SELL:
        player_id, resource_id, amount = struct.unpack_from('<bbb', data)
        return dict(player_id=player_id, resource_id=resource_id, amount=amount)
    if action_type == Action.DELETE:
        object_id, player_id = struct.unpack_from('<3x2I', data)
        return dict(player_id=player_id, object_ids=[object_id])
    if action_type == Action.TOWN_BELL:
        object_id, = struct.unpack_from('<3xI', data)
        return dict(object_ids=[object_id])
    if action_type == Action.WALL:
        selection_count, = struct.unpack_from('b', data)
        offset = len(data) - selection_count * 4

        if offset > 15:
            # In DE recordings all coordinates are prefixed with a zero.
            player_id, x_start, y_start, x_end, y_end, building_id, z1, const = struct.unpack_from(
                '<bxbxbxbxbx2h1i', data, offset=1)
        else:
            player_id, x_start, y_start, x_end, y_end, building_id, z1, const = struct.unpack_from(
                '<5bx2h1i', data, offset=1)

        object_ids = struct.unpack_from('<' + str(selection_count) + 'I', data[offset:])

        return dict(player_id=player_id,
                    x=x_start,
                    y=y_start,
                    x_end=x_end,
                    y_end=y_end,
                    building_id=building_id,
                    object_ids=object_ids)

    if action_type == Action.GAME:
        return dict(player_id=data[1], command_id=data[0])
    if action_type == Action.FLARE:
        x, y, player_id = struct.unpack_from('<19x2fb', data)
        return dict(player_id=player_id, x=x, y=y)
    if action_type == Action.REPAIR:
        target_id, *flags = struct.unpack_from('<3xI4b', data)
        offset = 0
        if check_flags(flags):
            offset = 4
        object_ids = struct.unpack_from('<' + str(data[0]) + 'I', data[7 + offset:])
        return dict(target_id=target_id, object_ids=list(object_ids))
    if action_type == Action.STOP:
        object_ids = struct.unpack_from('<x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids))
    if action_type == Action.GATE:
        object_id, = struct.unpack_from('<3xI', data)
        return dict(object_ids=[object_id])
    if action_type == Action.FOLLOW:
        object_ids = struct.unpack_from('<7x' + str(data[0]) + 'I', data)
        return dict(object_ids=object_ids)
    if action_type == Action.GUARD:
        object_ids = struct.unpack('<7x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids))
    if action_type == Action.ATTACK_GROUND:
        object_ids = []
        selected, x, y, *flags = struct.unpack_from('<b2x2f4b', data)
        offset = 0
        if check_flags(flags):
            offset = 4
        if selected > 0:
            object_ids = struct.unpack_from('<' + str(selected) + 'I', data[11 + offset:])
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.ADD_WAYPOINT:
        object_ids = []
        selected, x, y = struct.unpack_from('<xb2b', data)
        if selected > 0:
            object_ids = struct.unpack_from('<4x' + str(selected) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.DE_QUEUE:
        player_id, unit_id, amount, *object_ids = struct.unpack_from('<b4xhbx' + str(data[3]) + 'I', data)
        return dict(player_id=player_id, object_ids=object_ids, amount=amount, unit_id=unit_id)
    if action_type == Action.DE_ATTACK_MOVE:
        x, y, *object_ids = struct.unpack_from('<3xf36xf36x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids), x=x, y=y)
    if action_type == Action.DE_AUTOSCOUT:
        object_ids = struct.unpack_from('<x' + str(data[0]) + 'I', data)
        return dict(object_ids=list(object_ids))
    return dict()


def action(data, sequence=None):
    """Handle actions."""
    length, = struct.unpack('<I', data.read(4))
    action_id = int.from_bytes(data.read(1), 'little')
    action_bytes = data.read(length - 1)
    if sequence is None:
       sequence, = struct.unpack('<I', data.read(4))
    action_type = Action(action_id)
    if action_type == Action.POSTGAME:
        payload = dict(bytes=action_bytes + data.read())
    else:
        try:
            payload = parse_action(action_type, action_bytes)
        except struct.error:
            return Action.ERROR, {}
    payload["sequence"] = sequence
    return action_type, payload


def chat(data):
    """Handle chat."""
    _, length = struct.unpack('<II', data.read(8))
    msg = data.read(length)
    return msg


def start(data):
    """Handle start."""
    data.read(20)
    a, b, _ = struct.unpack('<III', data.read(12))
    if a != 0:  # AOC 1.0x
        data.seek(-12, 1)
    if b == 2:  # DE
        data.seek(-8, 1)


def save(data):
    """Handle saved chapter."""
    data.seek(-4, 1)
    pos = data.tell()
    length, _ = struct.unpack('<II', data.read(8))
    data.read(length - pos - 8)


def postgame(data):
    """Handle DE postgame."""
    data = io.BytesIO(data.read()[::-1])
    data.read(8)
    version, num_blocks = struct.unpack('>II', data.read(8))
    out = {}
    for _ in range(0, num_blocks):
        identifier, length = struct.unpack('>II', data.read(8))
        block = io.BytesIO(data.read(length)[::-1])
        postgame_type = Postgame(identifier)
        if postgame_type == Postgame.WORLD_TIME:
            out['world_time'] = struct.unpack('<I', block.read(4))[0]
        elif postgame_type == Postgame.LEADERBOARDS:
            num_leaderboards = struct.unpack('<I', block.read(4))[0]
            leaderboards = []
            for lb in range(0, num_leaderboards):
                leaderboard_id, unk = struct.unpack('<IH', block.read(6))
                num_players = struct.unpack('<I', block.read(4))[0]
                player_data = []
                for i in range(0, num_players):
                    player_num, rank, rating = struct.unpack('<3i', block.read(12))
                    player_data.append({
                        'number': player_num,
                        'rank': rank,
                        'rating': rating
                    })
                leaderboards.append({
                    'id': leaderboard_id,
                    'players': player_data
                })
            out['leaderboards'] = leaderboards
        else:
            raise RuntimeError("unparsed postgame block")
    return out


def meta(data):
    """Handle log meta."""
    try:
        first, = struct.unpack('<I', data.read(4))
        if first != 500:  # Not AOK
            data.read(4)
        data.read(20)
        a, b, _ = struct.unpack('<III', data.read(12))
        if a != 0:  # AOC 1.0x
            data.seek(-12, 1)
        if b == 2:  # DE
            data.seek(-8, 1)
    except struct.error:
        raise ValueError("insufficient meta received")


def operation(data):
    """Handle body operations."""
    try:
        op_id, = struct.unpack('<I', data.read(4))
        try:
            op_type = Operation(op_id)
        except ValueError:
            return Operation.SAVE, save(data)
        if op_type == Operation.ACTION:
            return op_type, action(data)
        if op_type == Operation.SYNC:
            return op_type, sync(data)
        if op_type == Operation.VIEWLOCK:
            return op_type, viewlock(data)
        if op_type == Operation.CHAT:
            return op_type, chat(data)
        if op_type == Operation.POSTGAME:
            return op_type, postgame(data)
    except struct.error:
        raise EOFError
    raise RuntimeError("unknown data received")



import unittest
from mgz.fast.header import parse
from mgz.util import Version

class TestFastUserPatch15(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('tests/recs/small.mgz', 'rb') as handle:
            cls.data = parse(handle)

    def test_version(self):
        self.assertEqual(self.data['version'], Version.USERPATCH15)

    def test_players(self):
        players = self.data.get('players')
        self.assertEqual(len(players), 3)
        self.assertEqual(players[0]['diplomacy'], [1, 4, 4, -1, -1, -1, -1, -1, -1])
        self.assertEqual(players[1]['diplomacy'], [0, 1, 4, -1, -1, -1, -1, -1, -1])
        self.assertEqual(players[2]['diplomacy'], [0, 4, 1, -1, -1, -1, -1, -1, -1])

    def test_map(self):
        self.assertEqual(self.data['scenario']['map_id'], 44)


class TestFastDE(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('tests/recs/de-13.34.aoe2record', 'rb') as handle:
            cls.data = parse(handle)

    def test_version(self):
        self.assertEqual(self.data['version'], Version.DE)

    def test_players(self):
        players = self.data.get('players')
        self.assertEqual(len(players), 3)

    def test_map(self):
        self.assertEqual(self.data['scenario']['map_id'], 9)
        self.assertEqual(self.data['lobby']['seed'], -1970180596)


class TestFastDEScenario(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('tests/recs/de-50.6-scenario.aoe2record', 'rb') as handle:
            cls.data = parse(handle)

    def test_version(self):
        self.assertEqual(self.data['version'], Version.DE)

    def test_players(self):
        players = self.data.get('players')
        self.assertEqual(len(players), 3)


class TestFastDEScenarioWithTriggers(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('tests/recs/de-50.6-scenario-with-triggers.aoe2record', 'rb') as handle:
            cls.data = parse(handle)

    def test_version(self):
        self.assertEqual(self.data['version'], Version.DE)

    def test_players(self):
        players = self.data.get('players')
        self.assertEqual(len(players), 3)


class TestFastHD(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('tests/recs/hd-5.8.aoe2record', 'rb') as handle:
            cls.data = parse(handle)

    def test_version(self):
        self.assertEqual(self.data['version'], Version.HD)

    def test_players(self):
        players = self.data.get('players')
        self.assertEqual(len(players), 7)

    def test_map(self):
        self.assertEqual(self.data['scenario']['map_id'], 0)



"""Chat Messages."""
import json
import logging
from enum import Enum

from mgz.fast import Age

LOGGER = logging.getLogger(__name__)
FEUDAL_AGE_MARKERS = [
    '봉건 시대',
    'Edad Feudal',
    '封建时代',
    'Feudalzeit',
    'Feodal Çağ',
    '封建時代',
    'Edad Feudal',
    '領主の時代',
    'Zaman Feudal',
    'Età feudale',
    'Feudal Age',
    'Thời phong kiến',
    'सामंतवादी युग',
    'Era Feudalna',
    'Idade Feudal',
    'Âge féodal',
    'Феодальная эпоха',
]
CASTLE_AGE_MARKERS = [
    '성주 시대',
    'Ed. Castillos',
    '城堡时代',
    'Ritterzeit',
    'Kale Çağı',
    '城堡時代',
    'Edad de los Castillos',
    '城主の時代',
    'Zaman Kastil',
    'Età dei castelli',
    'Castle Age',
    'Thời lâu đài',
    'परिवर्तन युग',
    'Era Zamków',
    'Idade dos Castelos',
    'Âge des châteaux',
    'Замковая эпоха',
]
IMPERIAL_AGE_MARKERS = [
    '왕정 시대',
    'Edad Imperial',
    '帝王时代',
    'Imperialzeit',
    'İmparatorluk Çağı',
    '帝王時代',
    'Edad Imperial',
    '帝王の時代',
    'Zaman Empayar',
    'Età imperiale',
    'Imperial Age',
    'Thời đế quốc',
    'साम्राज्यवादी युग',
    'Era Imperiów',
    'Idade Imperial',
    'Âge impérial',
    'Имперская эпоха',
]
AGE_MARKERS = [
    'advanced to the',
    'a progressé vers',
    '升级至',
    'avanzó a la',
    'đã phát triển lên',
    '시대로 발전했습니다',
    'vorangeschritten',
    'переход в',
    'avançou para a Idade',
    'wkroczyło w Erę',
    '升級至',
    'passaggio',
    'geÃ§ti',
    '進化し',
    'avançou para',
    'новую эпоху',
    'avanzó a Edad',
    'avanzado a la Edad',
    'ha raggiunto',
    'avanzó a Ed',
    'đã nâng cấp',
    'progressé vers',
    'wkracza do',
    'युग में उन्नत है।',   # hi
    'telah mara ke', # ms
    'geçti',         # tr
]
SAVE_MARKERS = [
    'Continuar con la partida en vez de guardar y salir',
    'Voto iniciado para guardar y salir del juego',
    'Chose to continue the game instead of save and exit',
    'Initiated vote to save and exit the game',
    'Vyber pokračovat ve hře místo ulo',
    'Escolha para continuar o jogo em vez de salvá-lo e fechá-lo',
    'Выберете Продолжить Игру вместо Сохранить и Выйти.',
    'Choisir pour continuer la partie au lieu d\'enregistrer et quitter.'
]


class Chat(Enum):
    """Chat types."""
    LADDER = 0
    VOOBLY = 1
    RATING = 2
    INJECTED = 3
    AGE = 4
    SAVE = 5
    MESSAGE = 6
    HELP = 7
    DISCARD = 8


def get_lobby_chat(header, encoding, diplomacy_type, players):
    """Get lobby chat."""
    chats = []
    if not hasattr(header.lobby, 'messages'):
        return chats
    for message in header.lobby.messages:
        if not message.message:
            continue
        try:
            chats.append(parse_chat(
                message.message, encoding, 0, players, diplomacy_type, origination='lobby'
            ))
        except UnicodeDecodeError:
            LOGGER.warning('could not decode lobby chat')
    return chats


def parse_chat(line, encoding, timestamp, players, diplomacy_type=None, origination='game'):
    """Initalize."""
    data = {
        'timestamp': timestamp,
        'origination': origination
    }
    try:
        line = line.strip(b'\x00').decode(encoding)
    except UnicodeDecodeError:
        data['type'] = Chat.DISCARD
        return data
    for save_marker in SAVE_MARKERS:
        if line.find(save_marker) > 0:
            data['type'] = Chat.SAVE
            return data
    if line.find('Voobly: Ratings provided') > 0:
        _parse_ladder(data, line)
    elif line.find('Voobly') == 3:
        _parse_voobly(data, line)
    elif line.find('<Rating>') > 0:
        _parse_rating(data, line)
    elif line.find('@#0<') == 0:
        _parse_injected(data, line)
    elif line.find('--') == 3:
        _parse_help(data, line)
    elif line.startswith('{"'):
        _parse_json(data, line, diplomacy_type)
    else:
        _parse_chat(data, line, players, diplomacy_type)

    if data['type'] is not Chat.DISCARD:
        for age_marker in AGE_MARKERS:
            if line.find(age_marker) > 0:
                if any(marker in line for marker in FEUDAL_AGE_MARKERS):
                    data['age'] = Age.FEUDAL_AGE
                if any(marker in line for marker in CASTLE_AGE_MARKERS):
                    data['age'] = Age.CASTLE_AGE
                if any(marker in line for marker in IMPERIAL_AGE_MARKERS):
                    data['age'] = Age.IMPERIAL_AGE
                if "age" in data:
                    data['type'] = Chat.AGE

    if not _validate(data, players):
        data['type'] = Chat.DISCARD
    return data


def _validate(data, players):
    """Chat messages can be bugged - check for invalid messages."""
    numbers = [p['number'] for p in players]
    if 'player_number' in data and data['player_number'] not in numbers:
        return False
    return True


def _parse_json(data, line, diplomacy_type):
    """Parse DE JSON chat."""
    payload = json.loads(line)
    if payload['messageAGP'] == '':
        # Use this condition to tell whether to ignore the chat. It might be from another game for example.
        data['type'] = Chat.DISCARD
        return

    audience = 'team'
    if payload['channel'] == 0:
        if diplomacy_type == '1v1':
            audience = 'all'
    elif payload['channel'] == 1:
        audience = 'all'
    data.update({
        'type': Chat.MESSAGE,
        'player_number': payload['player'],
        'message': payload['message'].strip(),
        'audience': audience
    })


def _parse_ladder(data, line):
    """Parse ladder from chat."""
    start = line.find("'") + 1
    end = line.find("'", start)
    data.update({
        'type': Chat.LADDER,
        'ladder': line[start:end]
    })


def _parse_voobly(data, line):
    """Parse voobly message from chat."""
    message = line[11:]
    data.update({
        'type': Chat.VOOBLY,
        'message': message
    })


def _parse_rating(data, line):
    """Parse rating from chat."""
    player_start = line.find('>') + 2
    player_end = line.find(':', player_start)
    player = line[player_start:player_end]
    rating = int(line[player_end + 2:len(line)])
    data.update({
        'type': Chat.RATING,
        'player': player,
        'rating': rating
    })


def _parse_injected(data, line):
    """Parse injected chat."""
    prefix = ''
    if line.find('<Team>') > 0:
        line = line.replace('<Team>', '', 1)
        prefix = ';'
    origination_start = line.find('<') + 1
    origination_end = line.find('>', origination_start)
    origination = line[origination_start:origination_end]
    name_end = line.find(':', origination_end)
    name = line[origination_end + 2:name_end]
    message = line[name_end + 2:]
    data.update({
        'type': Chat.INJECTED,
        'origination': origination.lower(),
        'name': name,
        'message': '{}{}'.format(prefix, message)
    })


def _parse_help(data, line):
    """Mark help-generated chat."""
    data['type'] = Chat.HELP


def _parse_chat(data, line, players, diplomacy_type):
    """Parse in-game chat message."""
    if not line or len(line) < 5:
        data['type'] = Chat.DISCARD
        return
    player_start = line.find('#') + 2
    if line[4] == ' ':
        player_start = line.find(' ') + 1
    player_end = line.find(':', player_start)
    player = line[player_start:player_end]
    if data['timestamp'] == 0:
        group = 'All'
    elif diplomacy_type == 'TG':
        group = 'Team'
    else:
        group = 'All'
    if player.find('>') > 0:
        group = player[1:player.find('>')]
        player = player[player.find('>') + 1:]
    if group.lower() in ['todos', 'всем', 'tous']:
        group = 'All'
    elif group.lower() in ['隊伍', 'squadra']:
        group = 'Team'
    message = line[player_end + 2:]
    number = None
    for player_h in players:
        if player_h['name'] in player:
            number = player_h['number']
    data.update({
        'type': Chat.MESSAGE,
        'player_number': number,
        'message': message.strip(),
        'audience': group.lower()
    })



"""Constants."""

# pylint: disable=too-many-lines

MODS = {
    1: 'Wololo Kingdoms',
    2: 'Portuguese Civilization Mod III',
    3: 'Age of Chivalry',
    4: 'Sengoku',
    7: 'Realms',
    101: 'King of the Hippo'
}

SPEEDS = {
    100: 'slow',
    150: 'standard',
    169: 'standard', # de
    178: 'standard', # up15
    200: 'fast',
    237: 'fast' # up15
}

DE_MAP_NAMES = {
    9: 'Arabia',
    10: 'Archipelago',
    11: 'Baltic',
    12: 'Black Forest',
    13: 'Coastal',
    14: 'Continental',
    15: 'Crater Lake',
    16: 'Fortress',
    17: 'Gold Rush',
    18: 'Highland',
    19: 'Islands',
    20: 'Mediterranean',
    21: 'Migration',
    22: 'Rivers',
    23: 'Team Islands',
    24: 'Full Random',
    25: 'Scandinavia',
    26: 'Mongolia',
    27: 'Yucatan',
    28: 'Salt Marsh',
    29: 'Arena',
    30: 'King of the Hill',
    31: 'Oasis',
    32: 'Ghost Lake',
    33: 'Nomad',
    49: 'Iberia',
    50: 'Britain',
    51: 'Mideast',
    52: 'Texas',
    53: 'Italy',
    54: 'Central America',
    55: 'France',
    56: 'Norse Lands',
    57: 'Sea of Japan (East Sea)',
    58: 'Byzantium',
    59: 'Custom',
    60: 'Random Land Map',
    62: 'Random Real World Map',
    63: 'Blind Random',
    65: 'Random Special Map',
    66: 'Random Special Map',
    67: 'Acropolis',
    68: 'Budapest',
    69: 'Cenotes',
    70: 'City of Lakes',
    71: 'Golden Pit',
    72: 'Hideout',
    73: 'Hill Fort',
    74: 'Lombardia',
    75: 'Steppe',
    76: 'Valley',
    77: 'MegaRandom',
    78: 'Hamburger',
    79: 'CtR Random',
    80: 'CtR Monsoon',
    81: 'CtR Pyramid Descent',
    82: 'CtR Spiral',
    83: 'Kilimanjaro',
    84: 'Mountain Pass',
    85: 'Nile Delta',
    86: 'Serengeti',
    87: 'Socotra',
    88: 'Amazon',
    89: 'China',
    90: 'Horn of Africa',
    91: 'India',
    92: 'Madagascar',
    93: 'West Africa',
    94: 'Bohemia',
    95: 'Earth',
    96: 'Canyons',
    97: 'Enemy Archipelago',
    98: 'Enemy Islands',
    99: 'Far Out',
    100: 'Front Line',
    101: 'Inner Circle',
    102: 'Motherland',
    103: 'Open Plains',
    104: 'Ring of Water',
    105: 'Snakepit',
    106: 'The Eye',
    107: 'Australia',
    108: 'Indochina',
    109: 'Indonesia',
    110: 'Strait of Malacca',
    111: 'Philippines',
    112: 'Bog Islands',
    113: 'Mangrove Jungle',
    114: 'Pacific Islands',
    115: 'Sandbank',
    116: 'Water Nomad',
    117: 'Jungle Islands',
    118: 'Holy Line',
    119: 'Border Stones',
    120: 'Yin Yang',
    121: 'Jungle Lanes',
    122: 'Alpine Lakes',
    123: 'Bogland',
    124: 'Mountain Ridge',
    125: 'Ravines',
    126: 'Wolf Hill',
    132: 'Antarctica',
    137: 'Custom Map Pool',
    139: 'Golden Swamp',
    140: 'Four Lakes',
    141: 'Land Nomad',
    142: 'Battle on Ice',
    143: 'El Dorado',
    144: 'Fall of Axum',
    145: 'Fall of Rome',
    146: 'Majapahit Empire',
    147: 'Amazon Tunnel',
    148: 'Coastal Forest',
    149: 'African Clearing',
    150: 'Atacama',
    151: 'Seize the Mountain',
    152: 'Crater',
    153: 'Crossroads',
    154: 'Michi',
    155: 'Team Moats',
    156: 'Volcanic Island'
}

MAP_NAMES = {
    9: 'Arabia',
    10: 'Archipelago',
    11: 'Baltic',
    12: 'Black Forest',
    13: 'Coastal',
    14: 'Continental',
    15: 'Crater Lake',
    16: 'Fortress',
    17: 'Gold Rush',
    18: 'Highland',
    19: 'Islands',
    20: 'Mediterranean',
    21: 'Migration',
    22: 'Rivers',
    23: 'Team Islands',
    24: 'Random',
    25: 'Scandinavia',
    26: 'Mongolia',
    27: 'Yucatan',
    28: 'Salt Marsh',
    29: 'Arena',
    30: 'King of the Hill',
    31: 'Oasis',
    32: 'Ghost Lake',
    33: 'Nomad',
    34: 'Iberia',
    35: 'Britain',
    36: 'Mideast',
    37: 'Texas',
    38: 'Italy',
    39: 'Central America',
    40: 'France',
    41: 'Norse Lands',
    42: 'Sea of Japan (East Sea)',
    43: 'Byzantinum',
    48: 'Blind Random',
    49: 'Acropolis',
    50: 'Budapest',
    51: 'Cenotes',
    52: 'City of Lakes',
    53: 'Golden Pit',
    54: 'Hideout',
    55: 'Hill Fort',
    56: 'Lombardia',
    57: 'Steppe',
    58: 'Valley',
    59: 'MegaRandom',
    60: 'Hamburger',
    61: 'CtR Random',
    62: 'CtR Monsoon',
    63: 'CtR Pyramid Descent',
    64: 'CtR Spiral',
    66: 'Acropolis',
    67: 'Budapest',
    68: 'Cenotes',
    69: 'City of Lakes',
    70: 'Golden Pit',
    71: 'Hideout',
    72: 'Hill Fort',
    73: 'Lombardia',
    74: 'Steppe',
    75: 'Valley',
    76: 'MegaRandom',
    77: 'Hamburger',
    78: 'CtR Random',
    79: 'CtR Monsoon',
    80: 'CtR Pyramid Descent',
    81: 'CtR Spiral',
    82: 'Kilimanjaro',
    83: 'Mountain Pass',
    84: 'Nile Delta',
    85: 'Serengeti',
    86: 'Socotra',
    87: 'Amazon',
    88: 'China',
    89: 'Horn of Africa',
    90: 'India',
    91: 'Madagascar',
    92: 'West Africa',
    93: 'Bohemia',
    94: 'Earth',
    95: 'Canyons',
    96: 'Enemy Archipelago',
    97: 'Enemy Islands',
    98: 'Far Out',
    99: 'Front Line',
    100: 'Inner Circle',
    101: 'Motherland',
    102: 'Open Plains',
    103: 'Ring of Water',
    104: 'Snakepit',
    105: 'The Eye',
    125: 'Ravines'
}

MAP_SIZES = {
    120: 'tiny',
    144: 'small',
    168: 'medium',
    200: 'normal',
    220: 'large',
    240: 'giant',
    255: 'maximum'
}

COMPASS = {
    'northwest': [1/3.0, 0],
    'southeast': [1/3.0, 2/3.0],
    'southwest': [0, 1/3.0],
    'northeast': [2/3.0, 1/3.0],
    'center': [1/3.0, 1/3.0],
    'west': [0, 0],
    'north': [2/3.0, 0],
    'east': [2/3.0, 2/3.0],
    'south': [0, 2/3.0]
}

VALID_BUILDINGS = [
    10, 12, 14, 18, 19, 20, 31, 32, 42, 45, 47, 49, 50, 51, 63, 64, 68, 70, 71, 72,
    78, 79, 82, 84, 85, 86, 87, 88, 90, 91, 101, 103, 104, 105, 109, 116, 117, 129,
    130, 131, 132, 133, 137, 141, 142, 153, 155, 199, 209, 210, 234, 235, 236, 276,
    331, 357, 463, 464, 465, 484, 487, 488, 490, 491, 498, 562, 563, 564, 565, 584,
    585, 586, 587, 597, 598, 617, 621, 659, 661, 665, 667, 669, 673, 674, 734, 789,
    790, 792, 793, 794, 796, 797, 798, 800, 801, 802, 804, 1189
] + [1553, 387, 110, 785, 1002] + [1021, 1187, 1251, 1665] + [946, 947, 886, 888, 881, 879, 938, 871] + [1665] # 5th Age, Realms, DE, Various, DE DLC1



import codecs
import unittest
from mgz.model import parse_match
from mgz.util import Version

class TestModel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('tests/recs/small.mgz', 'rb') as handle:
            cls.match = parse_match(handle)

    def test_match(self):
        self.assertEqual(self.match.speed, 'Standard')
        self.assertEqual(self.match.population, 200)
        self.assertEqual(self.match.diplomacy_type, '1v1')
        self.assertEqual(self.match.type, 'Random Map')
        self.assertEqual(self.match.dataset, 'Wololo Kingdoms')
        self.assertEqual(self.match.version, Version.USERPATCH15)

    def test_players(self):
        players = self.match.players
        self.assertEqual(len(players), 2)
        self.assertEqual(players[0].name, '[Heresy]LaaaaaN')
        self.assertEqual(players[0].number, 1)
        self.assertEqual(players[0].color, 'Red')
        self.assertEqual(players[0].civilization, 'Malians')
        self.assertFalse(players[0].winner)

    def test_objects(self):
        objects = self.match.players[0].objects
        self.assertEqual(len([o for o in objects if o.name == 'Villager']), 3)

    def test_map(self):
        self.assertEqual(self.match.map.name, 'KotD2 - Arabia')
        self.assertEqual(self.match.map.dimension, 120)
        self.assertEqual(self.match.map.size, 'Tiny')
        self.assertEqual(self.match.map.seed, -2119451194)
        self.assertEqual(len(self.match.map.tiles), 120 * 120)
        self.assertTrue(self.match.map.custom)

    def test_file(self):
        self.assertEqual(self.match.file.encoding, codecs.lookup('latin-1'))
        self.assertEqual(self.match.file.language, 'es')
        self.assertEqual(self.match.file.perspective.name, '[Heresy]LaaaaaN')



"""Determine diplomacy type."""


def get_diplomacy_type(teams, players):
    """Get type of diplomacy."""
    if len(teams) == 2 and len(players) > 2:
        return 'TG'
    elif len(players) == 2:
        return '1v1'
    elif (len(teams) == len(players) or len(teams) == 1) and len(players) > 2:
        return 'FFA'
    else:
        return 'Other'


def get_diplomacy_data(header, teams):
    """Compute diplomacy."""
    player_num = 0
    computer_num = 0
    for player in header.scenario.game_settings.player_info:
        if player.type == 'human':
            player_num += 1
        elif player.type == 'computer':
            computer_num += 1
    total_num = player_num + computer_num

    diplomacy = {}
    diplomacy['type'] = get_diplomacy_type(teams, header.initial.players[1:])
    team_sizes = sorted([len(team) for team in teams])
    diplomacy['team_size'] = 'v'.join([str(size) for size in team_sizes])
    if diplomacy['type'] == 'FFA':
        diplomacy['team_size'] = 'FFA'
    return diplomacy



"""Player stats."""

from construct import Embedded, Float32l, If, Struct, this, Array

from mgz.util import ModVersionAdapter, Version, find_version, find_save_version

# pylint: disable=invalid-name


# Snapshot taken at the start of the recorded game.
player_stats = "player_stats"/Struct(
    "resources"/Struct(
        "food"/Float32l,
        "wood"/Float32l,
        "stone"/Float32l,
        "gold"/Float32l,
    ),
    "headroom"/Float32l,
    "conversion_range"/Float32l,
    "current_age"/Float32l,
    "num_relics"/Float32l,
    "trade_bonus"/Float32l,
    "trade_goods"/Float32l,
    "trade_production"/Float32l,
    "pop_current"/Float32l,
    "decay"/Float32l,
    "discovery"/Float32l,
    "ruins"/Float32l,
    "meat"/Float32l,
    "berries"/Float32l,
    "fish"/Float32l,
    "u10"/Float32l,
    "total_units_created"/Float32l,
    "num_kills"/Float32l,
    "num_items_researched"/Float32l,
    "percent_map_explored"/Float32l,
    "castle_age"/Float32l,
    "imperial_age"/Float32l,
    "feudal_age"/Float32l,
    "u14"/Float32l,
    "convert_priests"/Float32l,
    "convert_buildings"/Float32l,
    "u17"/Float32l,
    "building_limit"/Float32l,
    "food_limit"/Float32l,
    "pop_max"/Float32l,
    "maintenance"/Float32l,
    "faith"/Float32l,
    "faith_recharge_rate"/Float32l,
    "farm_food_amount"/Float32l,
    "civilian_pop"/Float32l,
    "u23"/Float32l,
    "all_techs_achieved"/Float32l,
    "military_pop"/Float32l,
    "conversions"/Float32l,
    "num_wonders"/Float32l,
    "razings"/Float32l,
    "kill_ratio"/Float32l,
    "player_killed"/Float32l,
    "tribute_inefficiency"/Float32l,
    "gold_bonus_mining_productivity"/Float32l,
    "town_center_unavailable"/Float32l,
    "total_gold_gathered_2"/Float32l,
    "has_cartography"/Float32l,
    "houses_count"/Float32l,
    "monasteries"/Float32l,
    "total_resources_tributed"/Float32l,
    "hold_ruins"/Float32l,
    "hold_relics"/Float32l,
    "ore"/Float32l,
    "captured_unit"/Float32l,
    "dark_age"/Float32l,
    "trade_good_quality"/Float32l,
    "trade_market_level"/Float32l,
    "formations"/Float32l,
    "building_housing_rate"/Float32l,
    "gather_tax_rate"/Float32l,
    "gather_accumulator"/Float32l,
    "decay_rate"/Float32l,
    "allow_formations"/Float32l,
    "can_convert"/Float32l,
    "hit_points_killed"/Float32l,
    "p1_kills"/Float32l,
    "p2_kills"/Float32l,
    "p3_kills"/Float32l,
    "p4_kills"/Float32l,
    "p5_kills"/Float32l,
    "p6_kills"/Float32l,
    "p7_kills"/Float32l,
    "p8_kills"/Float32l,
    "convert_resistance"/Float32l,
    "trade_fee"/Float32l,
    "stone_mining_productivity"/Float32l,
    "num_units_queued"/Float32l,
    "num_units_making"/Float32l,
    "raider"/Float32l,
    "boarding_recharge_rate"/Float32l,
    "starting_villagers"/Float32l,
    "research_cost_mod"/Float32l,
    "research_time_mod"/Float32l,
    "convert_boats"/Float32l,
    "fish_trap_food"/Float32l,
    "heal_rate_modifier"/Float32l,
    "heal_range"/Float32l,
    "food_bonus"/Float32l,
    "wood_bonus"/Float32l,
    "stone_bonus"/Float32l,
    "gold_bonus"/Float32l,
    "raider_ability"/Float32l,
    "berserker_heal_timer"/Float32l,
    "dominant_sheep_control"/Float32l,
    "object_cost_summation"/Float32l,
    "research_cost_summation"/Float32l,
    "relic_income_summation"/Float32l,
    "trade_income_summation"/Float32l,
    "p1_tributed_to"/Float32l,
    "p2_tributed_to"/Float32l,
    "p3_tributed_to"/Float32l,
    "p4_tributed_to"/Float32l,
    "p5_tributed_to"/Float32l,
    "p6_tributed_to"/Float32l,
    "p7_tributed_to"/Float32l,
    "p8_tributed_to"/Float32l,
    "p1_unit_kill_worth"/Float32l,
    "p2_unit_kill_worth"/Float32l,
    "p3_unit_kill_worth"/Float32l,
    "p4_unit_kill_worth"/Float32l,
    "p5_unit_kill_worth"/Float32l,
    "p6_unit_kill_worth"/Float32l,
    "p7_unit_kill_worth"/Float32l,
    "p8_unit_kill_worth"/Float32l,
    "p1_num_razes"/Float32l,
    "p2_num_razes"/Float32l,
    "p3_num_razes"/Float32l,
    "p4_num_razes"/Float32l,
    "p5_num_razes"/Float32l,
    "p6_num_razes"/Float32l,
    "p7_num_razes"/Float32l,
    "p8_num_razes"/Float32l,
    "p1_raze_worth"/Float32l,
    "p2_raze_worth"/Float32l,
    "p3_raze_worth"/Float32l,
    "p4_raze_worth"/Float32l,
    "p5_raze_worth"/Float32l,
    "p6_raze_worth"/Float32l,
    "p7_raze_worth"/Float32l,
    "p8_raze_worth"/Float32l,
    "num_castles"/Float32l,
    "num_wonders"/Float32l,
    "kills_by_p1"/Float32l,
    "kills_by_p2"/Float32l,
    "kills_by_p3"/Float32l,
    "kills_by_p4"/Float32l,
    "kills_by_p5"/Float32l,
    "kills_by_p6"/Float32l,
    "kills_by_p7"/Float32l,
    "kills_by_p8"/Float32l,
    "razings_by_p1"/Float32l,
    "razings_by_p2"/Float32l,
    "razings_by_p3"/Float32l,
    "razings_by_p4"/Float32l,
    "razings_by_p5"/Float32l,
    "razings_by_p6"/Float32l,
    "razings_by_p7"/Float32l,
    "razings_by_p8"/Float32l,
    "value_killed_by_others"/Float32l,
    "value_razed_by_others"/Float32l,
    "kills"/Float32l,
    "losses"/Float32l,
    "p1_tribute_recvd"/Float32l,
    "p2_tribute_recvd"/Float32l,
    "p3_tribute_recvd"/Float32l,
    "p4_tribute_recvd"/Float32l,
    "p5_tribute_recvd"/Float32l,
    "p6_tribute_recvd"/Float32l,
    "p7_tribute_recvd"/Float32l,
    "p8_tribute_recvd"/Float32l,
    "current_unit_worth"/Float32l,
    "current_building_worth"/Float32l,
    "total_food_gathered"/Float32l,
    "total_wood_gathered"/Float32l,
    "total_stone_gathered"/Float32l,
    "total_gold_gathered"/Float32l,
    "total_kill_and_raze_worth"/Float32l,
    "total_tribute_recvd"/Float32l,
    "total_value_of_razings"/Float32l,
    "castle_high"/Float32l,
    "wonder_high"/Float32l,
    "total_tribute_sent"/Float32l,
    "convert_min_adj"/Float32l,
    "convert_max_adj"/Float32l,
    "convert_resist_min_adj"/Float32l,
    "convert_resist_max_adj"/Float32l,
    "convert_building_min"/Float32l,
    "convert_building_max"/Float32l,
    "convert_building_chance"/Float32l,
    "spies"/Float32l,
    "value_wonders_castles"/Float32l,
    "food_score"/Float32l,
    "wood_score"/Float32l,
    "stone_score"/Float32l,
    "gold_score"/Float32l,
    Embedded(If(lambda ctx: find_save_version(ctx) >= 11.76, Struct(
        "wood_bonus0"/Float32l,
        "food_bonus0"/Float32l,
        "relic_rate"/Float32l,
        "heresy"/Float32l,
        "theocracy"/Float32l,
        "crenellations"/Float32l,
        "construction_rate_mod"/Float32l,
        "hun_wonder_bonus"/Float32l,
        "spies_discount"/Float32l,
    ))),
    Embedded(If(lambda ctx: find_version(ctx) in (Version.DE, Version.HD), Struct(
        Array(this._._.num_header_data - 198, Float32l)
    ))),
    Embedded(If(lambda ctx: find_save_version(ctx) >= 63, Struct(
        Array(this._._.num_header_data, Float32l)
    ))),
    Embedded(If(lambda ctx: find_version(ctx) in [Version.USERPATCH15, Version.MCP], Struct(
        ModVersionAdapter("mod"/Float32l),
        Array(6, Float32l),
        "trickle_food"/Float32l,
        "trickle_wood"/Float32l,
        "trickle_stone"/Float32l,
        "trickle_gold"/Float32l,
        "reveal_enemy_tcs"/Float32l,
        "reveal_gaia_class_1"/Float32l,
        "reveal_gaia_class_2"/Float32l,
        Array(28, Float32l)
    ))),
    Embedded(If(this._._._._.version == Version.MCP, Struct(
        Array(65, Float32l)
    )))
)



"""Convert parsed data into object-oriented model."""

import codecs
import collections
import _hashlib
import hashlib
from datetime import timedelta, datetime
from enum import Enum

import dataclasses

from mgz import fast
from mgz.reference import get_consts, get_dataset
from mgz.fast import Action as ActionEnum
from mgz.fast.header import parse
from mgz.model.definitions import *
from mgz.model.inputs import Inputs
from mgz.common.chat import parse_chat, Chat as ChatEnum
from mgz.common.diplomacy import get_diplomacy_type
from mgz.common.map import get_map_data
from mgz.util import Version


TC_IDS = [71, 109, 141, 142]
AI_ACTIONS = [ActionEnum.AI_ORDER]


def enrich_action(action, action_data, dataset, consts):
    """Enrich action data with lookups."""
    if 'x' in action_data and 'y' in action_data and action_data['x'] >=0 and action_data['y'] >= 0:
        if action.type != fast.Action.SPECIAL or ('target_id' in action_data and action_data['target_id'] > 0):
            action.position = Position(action_data['x'], action_data['y'])
            del action.payload['x']
            del action.payload['y']
    if 'technology_id' in action_data:
        action.payload['technology'] = dataset['technologies'].get(str(action_data['technology_id']))
    if 'formation_id' in action_data:
        action.payload['formation'] = consts['formations'].get(str(action_data['formation_id']))
    if 'stance_id' in action_data:
        action.payload['stance'] = consts['stances'].get(str(action_data['stance_id']))
    if 'building_id' in action_data:
        action.payload['building'] = dataset['objects'].get(str(action_data['building_id']))
    if 'unit_id' in action_data:
        action.payload['unit'] = dataset['objects'].get(str(action_data['unit_id']))
    if 'command_id' in action_data:
        action.payload['command'] = consts['commands'].get(str(action_data['command_id']))
    if 'order_id' in action_data:
        action.payload['order'] = consts['orders'].get(str(action_data['order_id']))
    if 'resource_id' in action_data:
        action.payload['resource'] = consts['resources'].get(str(action_data['resource_id']))


def get_difficulty(data):
    if data['version'] is Version.HD:
        return data['hd']['difficulty_id']
    elif data['version'] is Version.DE:
        return data['de']['difficulty_id']
    return data['scenario']['difficulty_id']


def get_team_together(data):
    if data['version'] is Version.DE:
        return data['de']['team_together']
    return None


def get_lock_speed(data):
    if data['version'] is Version.DE:
        return data['de']['lock_speed']
    return None


def get_all_technologies(data):
    if data['version'] is Version.DE:
        return data['de']['all_technologies']
    return None

def get_starting_age(data):
    if data['version'] is Version.DE:
        return data['de']['starting_age_id']
    return None


def get_map_id(data):
    if data['version'] is Version.HD:
        return data['hd']['map_id']
    if data['version'] is Version.DE:
        return data['de']['rms_map_id']
    return data['scenario']['map_id']


def get_hash(data):
    if data['version'] is Version.DE:
        return data['de']['hash']
    return None


def parse_match(handle):
    """Parse a match.

    This is one big function because the dependency graph between
    the variables is dense.
    """

    data = parse(handle)
    body_pos = handle.tell() - 4 # log version
    consts = get_consts()

    dataset_id, dataset = get_dataset(data['version'], data['mod'])
    map_id = get_map_id(data)
    try:
            map_data, encoding, language = get_map_data(
            map_id,
            data['scenario']['instructions'],
            data['map']['dimension'],
            data['version'],
            dataset_id,
            dataset,
            data['map']['tiles'],
            de_seed=data['lobby']['seed']
        )
    except ValueError as e:
        raise RuntimeError(f"could not get map data: {e}")

    # Handle DE-specific data
    rated = None
    if data['de']:
        de_players = {player['number']: player for player in data['de']['players']}
        lobby = data['de']['lobby']
        guid = data['de']['guid']
        rated = data['de']['rated']
    else:
        de_players = dict()
        lobby = None
        guid = None

    # Parse gaia objects
    gaia = [
        Object(
            dataset['objects'].get(str(obj['object_id'])),
            obj['class_id'],
            obj['object_id'],
            obj['instance_id'],
            obj['index'],
            Position(obj['position']['x'], obj['position']['y'])
        )
        for obj in data['players'][0]['objects']
    ]

    inputs = Inputs({o.instance_id:o.name for o in gaia})

    # Parse players
    players = dict()
    allies = dict()
    for player in data['players'][1:]:
        allies[player['number']] = set([player['number']])
        for i, stance in enumerate(player['diplomacy']):
            if stance == 2:
                allies[player['number']].add(i)
        de_player = de_players.get(player['number'])
        if de_player:
            player.update(de_player)
        pos_x = None
        pos_y = None
        for obj in player['objects']:
            if obj['object_id'] in TC_IDS:
                pos_x = obj['position']['x']
                pos_y = obj['position']['y']
        players[player['number']] = Player(
            player['number'],
            player['name'].decode(encoding),
            consts['player_colors'][str(player['color_id'])],
            player['color_id'],
            dataset['civilizations'][str(player['civilization_id'])]['name'],
            player['civilization_id'],
            Position(pos_x, pos_y),
            [
                Object(
                    dataset['objects'].get(str(obj['object_id'])),
                    obj['class_id'],
                    obj['object_id'],
                    obj['instance_id'],
                    obj['index'],
                    Position(obj['position']['x'], obj['position']['y'])
                )
                for obj in player['objects']
            ],
            player.get('profile_id'),
            [],
            player.get('prefer_random'),
        )

    # Assign teams
    if de_players:
        by_team = collections.defaultdict(list)
        for number, player in de_players.items():
            if player['team_id'] > 1:
                by_team[player['team_id']].append(number)
            elif player['team_id'] == 1:
                by_team[number + 9].append(number)
        team_ids = by_team.values()
    else:
        team_ids = set([frozenset(s) for s in allies.values()])
    teams = []
    for team in team_ids:
        t = [players[x] for x in team]
        for x in team:
            players[x].team = t
            players[x].team_id = team
        teams.append(t)

    # Compute diplomacy
    diplomacy_type = get_diplomacy_type(teams, players)

    # Extract lobby chat
    pd = [dict(name=p.name, number=n) for n, p in players.items()]
    chats = []
    for c in data['lobby']['chat']:
        chat = parse_chat(c, encoding, 0, pd, diplomacy_type, 'lobby')
        if chat['type'] == ChatEnum.DISCARD or chat['player_number'] not in players:
            continue
        chats.append(Chat(
            timedelta(milliseconds=chat['timestamp']),
            chat['message'],
            chat['origination'],
            chat['audience'],
            players[chat['player_number']]
        ))
        inputs.add_chat(chats[-1])

    # Parse player actions
    fast.meta(handle)
    timestamp = 0
    resigned = []
    actions = []
    viewlocks = []
    uptimes = []
    eapm = collections.Counter()
    last_viewlock = None
    while True:
        try:
            op_type, op_data = fast.operation(handle)
            if op_type is fast.Operation.SYNC:
                timestamp += op_data[0]
                if op_data[2]:
                    stat_row = op_data[2]
                    for player in players.values():
                        if player.number not in stat_row:
                            continue
                        stats = stat_row[player.number]
                        player.timeseries.append(TimeseriesRow(
                            timestamp=timedelta(milliseconds=stat_row['current_time']),
                            total_resources=stats['total_res'],
                            total_objects=stats['obj_count']
                        ))
            elif op_type is fast.Operation.VIEWLOCK:
                if op_data == last_viewlock:
                    continue
                viewlock = Viewlock(timedelta(milliseconds=timestamp), Position(*op_data), players[data['metadata']['owner_id']])
                viewlocks.append(viewlock)
                last_viewlock = op_data
            elif op_type is fast.Operation.CHAT:
                chat = parse_chat(op_data, encoding, timestamp, pd, diplomacy_type, 'game')
                if chat['type'] == ChatEnum.MESSAGE:
                    chats.append(Chat(
                        timedelta(milliseconds=chat['timestamp'] + data['map']['restore_time']),
                        chat['message'],
                        chat['origination'],
                        chat['audience'],
                        players[chat['player_number']]
                    ))
                    inputs.add_chat(chats[-1])
                if chat['type'] == ChatEnum.AGE:
                    uptimes.append(
                        Uptime(
                            timedelta(milliseconds=chat['timestamp'] + data['map']['restore_time']),
                            chat['age'],
                            players.get(chat['player_number']),
                        )
                    )
            elif op_type is fast.Operation.ACTION:
                action_type, action_data = op_data
                action = Action(timedelta(milliseconds=timestamp), action_type, action_data)
                if action_type is fast.Action.RESIGN and action_data['player_id'] in players:
                    resigned.append(players[action_data['player_id']])
                if 'player_id' in action_data and action_data['player_id'] in players:
                    if action_type not in AI_ACTIONS:
                        eapm[action_data['player_id']] += 1
                    action.player = players[action_data['player_id']]
                    del action.payload['player_id']
                enrich_action(action, action_data, dataset, consts)
                actions.append(action)
                inputs.add_action(action)
            elif op_type is fast.Operation.POSTGAME:
                by_number = {x["number"]: x["rating"] for x in op_data["leaderboards"][0]["players"]}
                for player in players.values():
                    player.rate_snapshot = by_number.get(player.number - 1)
        except EOFError:
            break

    # Compute winner(s)
    for team in teams:
        winner = not any([player for player in team if player in resigned])
        if resigned:
            for player in team:
                player.winner = winner

    # Compute eAPM
    for player_id, action_count in eapm.items():
        players[player_id].eapm = int(round(eapm[player_id] / ((timestamp/1000)/60)))

    handle.seek(body_pos)
    file_bytes = handle.read()
    file_size = body_pos + 4 + len(file_bytes)
    file_hash = hashlib.sha1(file_bytes).hexdigest()
    return Match(
        list(players.values()),
        teams,
        gaia,
        Map(
            map_id,
            map_data['name'],
            map_data['dimension'],
            consts['map_sizes'].get(str(map_data['dimension'])),
            map_data['custom'],
            map_data['seed'],
            data['de']['rms_mod_id'] if data['version'] is Version.DE and map_data['custom'] else None,
            map_data['name'].startswith('ZR@'),
            map_data['modes'],
            [
                Tile(
                    tile['terrain_id'],
                    tile['elevation'],
                    Position(tile['x'], tile['y'])
                ) for tile in map_data['tiles']
            ]
        ),
        File(
            codecs.lookup(encoding),
            language,
            file_hash,
            file_size,
            data['device'],
            players[data['metadata']['owner_id']],
            viewlocks
        ),
        data['map']['restore_time'] > 0,
        timedelta(milliseconds=data['map']['restore_time']),
        consts['speeds'][str(int(round(data['metadata']['speed'], 2) * 100))],
        int(round(data['metadata']['speed'], 2) * 100),
        data['metadata']['cheats'],
        data['lobby']['lock_teams'],
        data['lobby']['population'],
        chats,
        guid,
        lobby,
        rated,
        dataset['dataset']['name'],
        consts['game_types'][str(data['lobby']['game_type_id'])],
        data['lobby']['game_type_id'],
        consts['map_reveal_choices'][str(data['lobby']['reveal_map_id'])],
        data['lobby']['reveal_map_id'],
        consts['difficulties'].get(str(get_difficulty(data))),
        get_difficulty(data),
        consts['starting_ages'].get(str(get_starting_age(data))),
        get_starting_age(data),
        get_team_together(data),
        get_lock_speed(data),
        get_all_technologies(data),
        True if data['version'] is Version.DE else None,
        timedelta(milliseconds=timestamp + data['map']['restore_time']),
        diplomacy_type,
        bool(resigned),
        dataset_id,
        data['version'],
        data['game_version'],
        data['save_version'],
        data['log_version'],
        data['de']['build'] if data['version'] is Version.DE else None,
        datetime.fromtimestamp(data['de']['timestamp']) if data['version'] is Version.DE and data['de']['timestamp'] else None,
        timedelta(seconds=data['de']['spec_delay']) if data['version'] is Version.DE else None,
        data['de']['allow_specs'] if data['version'] is Version.DE else None,
        data['de']['hidden_civs'] if data['version'] is Version.DE else None,
        data['de']['visibility_id'] == 2 if data['version'] is Version.DE else None,
        get_hash(data),
        actions,
        inputs.inputs,
        uptimes
    )


def serialize(obj):
    """Serialize model.

    Returns a nested datastructure with no circular references,
    appropriate for dumping to JSON, YAML, etc.
    """
    seen = set()

    def impl(obj):
        """Recursive serialization implementation."""
        if dataclasses.is_dataclass(obj) and isinstance(obj, collections.abc.Hashable):
            if obj in seen:
                return hash(obj)
            seen.add(obj)
        if type(obj) is list:
            return [v for v in [impl(o) for o in obj] if v is not None]
        elif type(obj) is dict:
            return {k:v for k, v in {f:impl(d) for f, d in obj.items()}.items() if v is not None}
        elif dataclasses.is_dataclass(obj):
            return {k:v for k, v in {f.name:impl(getattr(obj, f.name)) for f in dataclasses.fields(obj)}.items() if v is not None}
        elif isinstance(obj, (codecs.CodecInfo, Enum)):
            return obj.name
        elif isinstance(obj, timedelta):
            return str(obj)
        elif isinstance(obj, datetime):
            return str(obj)
        elif isinstance(obj, bytes):
            return None
        elif isinstance(obj, _hashlib.HASH):
            return obj.hexdigest()
        else:
            return obj

    return impl(obj)



"""Determine player data."""

from mgz.summary.objects import TC_IDS
from collections import defaultdict


def enrich_de_player_data(players, extraction):
    """Enrich player data with extracted data."""
    last_records = {}
    research_count = defaultdict(int)
    castles = defaultdict(int)
    wonders = defaultdict(int)
    age_times = {}

    for player in players:
        last_records[player['number']] = None
        age_times[player['number']] = {}

    for record in reversed(extraction['timeseries']):
        if not last_records[record['player_number']]:
            last_records[record['player_number']] = record
        if all(last_records.values()):
            break

    for obj in extraction['objects']:
        if obj['initial_object_id'] == 82:
            castles[obj['initial_player_number']] += 1
        elif obj['initial_object_id'] == 276:
            wonders[obj['initial_player_number']] += 1

    for research in extraction['research']:
        if not research['finished']:
            continue
        if research['technology_id'] in [101, 102, 103]:
            age_times[research['player_number']][research['technology_id']] = int(research['started'] / 1000)
        research_count[research['player_number']] += 1

    for player in players:
        record = last_records[player['number']]
        player.update(dict(
            score=record['total_score'],
            winner=player['number'] in extraction['winners']
        ))
        player['achievements']['military'].update(dict(
            score=record['military_score'],
            units_killed=record['kills'],
            units_lost=record['deaths'],
            buildings_lost=record['buildings_lost'],
            buildings_razed=record['razes'],
            units_converted=record['converted'],
            hit_points_killed=record['hit_points_killed'],
            hit_points_razed=record['hit_points_razed']
        ))
        player['achievements']['economy'].update(dict(
            score=record['economy_score'],
            food_collected=record['total_food'],
            wood_collected=record['total_wood'],
            stone_collected=record['total_stone'],
            gold_collected=record['total_gold'],
            tribute_sent=record['tribute_sent'],
            tribute_received=record['tribute_received'],
            trade_gold=record['trade_profit'],
            relic_gold=record['relic_gold']
        ))
        player['achievements']['society'].update(dict(
            score=record['society_score'],
            total_relics=record['relics_captured'],
            total_castles=castles[player['number']],
            total_wonders=wonders[player['number']],
            villager_high=record['villager_high']
        ))
        player['achievements']['technology'].update(dict(
            score=record['technology_score'],
            explored_percent=record['percent_explored'],
            research_count=research_count[player['number']],
            feudal_time=age_times[player['number']].get(101),
            castle_time=age_times[player['number']].get(102),
            imperial_time=age_times[player['number']].get(103)
        ))


def ach(structure, fields):
    """Get field from achievements structure."""
    field = fields.pop(0)
    if structure:
        if hasattr(structure, field):
            structure = getattr(structure, field)
            if not fields:
                return structure
            return ach(structure, fields)
    return None


def get_achievements(postgame, name):
    """Get achievements for a player.

    Must match on name, not index, since order is not always the same.
    """
    if not postgame:
        return None
    for achievements in postgame.achievements:
        # achievements player name can be shorter
        if name.startswith(achievements.player_name.replace(b'\x00', b'')):
            return achievements
    return None


def guess_winner(teams, resigned, i):
    """Guess if a player won.

    Find what team the player was on. If anyone
    on their team resigned, assume the player lost.

    If there were no resignations, the game ended
    by some other condition (wonder, timelimit, etc),
    assuming it did complete.
    """
    for team in teams:
        if i not in team:
            continue
        for j in team:
            if j in resigned:
                return False
    if len(resigned) > 0:
        return True
    return None


def get_civilization(header, index):
    """Get civilization ID."""
    if header.save_version >= 20.06:
        return header.de.players[index].civ_id
    return header.initial.players[index + 1].attributes.civilization


def get_color(header, index):
    """Get color ID."""
    if header.save_version >= 20.06:
        return header.de.players[index].color_id
    return header.initial.players[index + 1].attributes.player_color


def get_position(header, index):
    """Get position."""
    if header.save_version >= 20.06:
        for obj in header.initial.players[index + 1].objects:
            if obj.object_type in TC_IDS:
                return (obj.x, obj.y)
    attr = header.initial.players[index + 1].attributes
    return (attr.camera_x, attr.camera_y)


def get_prefer_random(header, index):
    if header.save_version >= 20.06:
        return header.de.players[index].prefer_random
    return None


def get_players_data(header, postgame, teams, resigned, cheaters, profile_ids, ratings, encoding, eapm): # pylint: disable=too-many-arguments, too-many-locals
    """Get basic player info."""
    out = []
    for i, player in enumerate(header.initial.players[1:]):
        achievements = get_achievements(postgame, player.attributes.player_name)
        if achievements:
            winner = achievements.victory
        else:
            winner = guess_winner(teams, resigned, i + 1)
        feudal_time = ach(achievements, ['technology', 'feudal_time_int'])
        castle_time = ach(achievements, ['technology', 'castle_time_int'])
        imperial_time = ach(achievements, ['technology', 'imperial_time_int'])
        name = player.attributes.player_name.decode(encoding)
        out.append({
            'id': i,
            'name': name,
            'civilization': get_civilization(header, i),
            'human': header.scenario.game_settings.player_info[i + 1].type == 'human',
            'number': i + 1,
            'color_id': get_color(header, i),
            'team_id': header.lobby.teams[i] - 2,
            'winner': winner,
            'mvp': ach(achievements, ['mvp']),
            'score': ach(achievements, ['total_score']),
            'position': get_position(header, i),
            'rate_snapshot': ratings.get(name),
            'user_id': profile_ids.get(i + 1),
            'cheater': (i + 1) in cheaters,
            'prefer_random': get_prefer_random(header, i),
            'eapm': eapm.get(i + 1),
            'achievements': {
                'military': {
                    'score': ach(achievements, ['military', 'score']),
                    'units_killed': ach(achievements, ['military', 'units_killed']),
                    'hit_points_killed': ach(achievements, ['military', 'hit_points_killed']),
                    'units_lost': ach(achievements, ['military', 'units_lost']),
                    'buildings_razed': ach(achievements, ['military', 'buildings_razed']),
                    'hit_points_razed': ach(achievements, ['military', 'hit_points_razed']),
                    'buildings_lost': ach(achievements, ['military', 'buildings_lost']),
                    'units_converted': ach(achievements, ['military', 'units_converted'])
                },
                'economy': {
                    'score': ach(achievements, ['economy', 'score']),
                    'food_collected': ach(achievements, ['economy', 'food_collected']),
                    'wood_collected': ach(achievements, ['economy', 'wood_collected']),
                    'stone_collected': ach(achievements, ['economy', 'stone_collected']),
                    'gold_collected': ach(achievements, ['economy', 'gold_collected']),
                    'tribute_sent': ach(achievements, ['economy', 'tribute_sent']),
                    'tribute_received': ach(achievements, ['economy', 'tribute_received']),
                    'trade_gold': ach(achievements, ['economy', 'trade_gold']),
                    'relic_gold': ach(achievements, ['economy', 'relic_gold'])
                },
                'technology': {
                    'score': ach(achievements, ['technology', 'score']),
                    'feudal_time': feudal_time if feudal_time and feudal_time > 0 else None,
                    'castle_time': castle_time if castle_time and castle_time > 0 else None,
                    'imperial_time': imperial_time if imperial_time and imperial_time > 0 else None,
                    'explored_percent': ach(achievements, ['technology', 'explored_percent']),
                    'research_count': ach(achievements, ['technology', 'research_count']),
                    'research_percent': ach(achievements, ['technology', 'research_percent'])
                },
                'society': {
                    'score': ach(achievements, ['society', 'score']),
                    'total_wonders': ach(achievements, ['society', 'total_wonders']),
                    'total_castles': ach(achievements, ['society', 'total_castles']),
                    'total_relics': ach(achievements, ['society', 'relics_captured']),
                    'villager_high': ach(achievements, ['society', 'villager_high'])
                }
            }
        })
    return out



"""Objects."""
from collections import defaultdict


TC_IDS = [71, 109, 141, 142]
STONE_WALL_ID = 117
PALISADE_WALL_ID = 72
CLIFFS = [264, 265, 266, 267, 268, 269, 270, 271, 272, 273]


def get_objects_data(header):
    """Get starting objects data."""
    objects = []
    tcs = defaultdict(int)
    stone_walls = {}
    palisade_walls = {}
    annexes = set()
    for player in header.initial.players:
        sleeping = []
        for o in player.sleeping_objects:
            if player.type == 2 and o.object_type not in CLIFFS:
                continue
            sleeping.append(o)
        for o in player.objects + sleeping:
            # static, moving, unit, and buildings only
            if o.type not in [10, 30, 70, 80]:
                continue
            # count town centers
            if o.object_type in TC_IDS:
                tcs[o.player_id] += 1
            # check for arena
            elif o.object_type == STONE_WALL_ID:
                stone_walls[o.player_id] = True
            # check for hideout
            elif o.object_type == PALISADE_WALL_ID:
                palisade_walls[o.player_id] = True
            # skip annexes
            if o.type == 80 and o.object_id in annexes:
                continue
            objects.append(dict(
                instance_id=o.object_id,
                object_id=o.object_type,
                class_id=o.type,
                player_number=o.player_id if o.player_id > 0 else None,
                x=o.x,
                y=o.y
            ))

    return dict(
        objects=objects,
        tcs=max(tcs.values()) if len(tcs.values()) > 0 else None,
        stone_walls=bool(stone_walls) and all(stone_walls.values()),
        palisade_walls=bool(palisade_walls) and all(palisade_walls.values())
    )



"""Initial."""

# pylint: disable=invalid-name,no-name-in-module

from construct import (Array, Byte, Embedded, Flag, Float32l, If, Int16ul, Int32sl,
                       Int32ul, Padding, Struct, Tell, this, Bytes, Const, IfThenElse)

from mgz.enums import MyDiplomacyEnum, TheirDiplomacyEnum
from mgz.header.objects import existing_object
from mgz.header.playerstats import player_stats
from mgz.util import Find, GotoObjectsEnd, RepeatUpTo, Version, find_save_version, find_version

# Player attributes.
attributes = "attributes"/Struct(
    Array(lambda ctx: ctx._._._.replay.num_players, TheirDiplomacyEnum("their_diplomacy"/Byte)),
    Array(lambda ctx: ctx._._._.replay.num_players if find_save_version(ctx) >= 61.5 else 9, MyDiplomacyEnum("my_diplomacy"/Int32sl)),
    "allied_los"/Int32ul,
    "allied_victory"/Flag,
    "player_name_length"/Int16ul,
    "player_name"/Bytes(this.player_name_length - 1),
    Padding(1), # 0x00
    Padding(1), # 0x16
    "num_header_data"/Int32ul,
    Padding(1), # 0x21
    player_stats,
    Padding(1),
    "camera_x"/Float32l,
    "camera_y"/Float32l,
    "end_of_camera"/Tell,
    Embedded(If(this._._._.version != Version.AOK, Struct(
        "num_saved_views"/Int32sl,
        # present in resumed games
        If(lambda ctx: ctx.num_saved_views > 0, Array(
            lambda ctx: ctx.num_saved_views, "saved_view"/Struct(
                "camera_x"/Float32l,
                "camera_y"/Float32l
            )
        )),
    ))),
    "spawn_location"/Struct(
        "x"/Int16ul,
        "y"/Int16ul
    ),
    "culture"/Byte,
    "civilization"/Byte,
    "game_status"/Byte,
    "resigned"/Flag,
    Padding(1),
    "player_color"/Byte,
    Padding(1),
)


# Initial state of players, including Gaia.
player = "players"/Struct(
    "type"/Byte,
    "unk"/Byte,
    attributes,
    "end_of_attr"/Tell,
    "start_of_objects"/Find([b'\x0b\x00.\x00\x00\x00\x02\x00\x00'], None),
    Embedded(IfThenElse(lambda ctx: ctx._.restore_time == 0,
        Struct(
            "objects"/RepeatUpTo(b'\x00', existing_object),
            Const(b'\x00\x0b'),
            # Skip Gaia trees for performance reasons
            Embedded(IfThenElse(lambda ctx: ctx._.type != 2 or find_version(ctx).value < 10,
                Struct(
                    "s_size"/Int32ul,
                    "s_grow"/Int32ul,
                    "sleeping_objects"/RepeatUpTo(b'\x00', existing_object),
                    Const(b'\x00\x0b'),
                    "d_size"/Int32ul,
                    "d_grow"/Int32ul,
                    "doppleganger_objects"/RepeatUpTo(b'\x00', existing_object),
                    Const(b'\x00\x0b')
                ),
                Struct(
                    "sleeping_objects"/Array(0, existing_object),
                    "doppleganger_objects"/Array(0, existing_object)
                )
            ))
        ),
        Struct(
            "objects"/Array(0, existing_object),
            "sleeping_objects"/Array(0, existing_object),
            "doppleganger_objects"/Array(0, existing_object)
        )
    )),
    "end_of_objects"/GotoObjectsEnd()
)


# Initial state.
initial = "initial"/Struct(
    "restore_time"/Int32ul, # zero for non-restored
    "num_particles"/Int32ul,
    "particles"/Bytes(lambda ctx: ctx.num_particles * 27),
    "identifier"/Int32ul,
    Array(lambda ctx: ctx._.replay.num_players, player),
    Padding(21),
)



"""Setup."""
from setuptools import setup, find_packages

setup(
    name='mgz',
    version='1.8.42',
    description='Parse Age of Empires 2 recorded games.',
    url='https://github.com/happyleavesaoc/aoc-mgz/',
    license='MIT',
    author='happyleaves',
    author_email='happyleaves.tfr@gmail.com',
    packages=find_packages(),
    install_requires=[
        'aocref>=2.0.28',
        'construct==2.8.16',
        'dataclasses==0.8; python_version < "3.7"',
        'tabulate>=0.9.0',
    ],
    entry_points = {
        'console_scripts': ['mgz=mgz.cli:main'],
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ]
)



"""Fast body parser enumerations."""

from enum import Enum

class Operation(Enum):
    """Operation types."""
    ACTION = 1
    SYNC = 2
    VIEWLOCK = 3
    CHAT = 4
    START = 5
    POSTGAME = 6
    SAVE = 7


class Action(Enum):
    """Action types."""
    ERROR = -1
    ORDER = 0
    STOP = 1
    WORK = 2
    MOVE = 3
    CREATE = 4
    ADD_ATTRIBUTE = 5
    GIVE_ATTRIBUTE = 6
    AI_ORDER = 10
    RESIGN = 11
    SPECTATE = 15
    ADD_WAYPOINT = 16
    STANCE = 18
    GUARD = 19
    FOLLOW = 20
    PATROL = 21
    FORMATION = 23
    SAVE = 27
    GROUP_MULTI_WAYPOINTS = 31
    CHAPTER = 32
    DE_ATTACK_MOVE = 33
    HD_UNKNOWN_34 = 34
    DE_RETREAT = 35
    DE_UNKNOWN_37 = 37
    DE_AUTOSCOUT = 38
    DE_UNKNOWN_39 = 39
    DE_UNKNOWN_40 = 40
    DE_TRANSFORM = 41
    RATHA_ABILITY = 43
    DE_107_A = 44
    DE_MULTI_GATHERPOINT = 45
    AI_COMMAND = 53
    DE_UNKNOWN_80 = 80
    MAKE = 100
    RESEARCH = 101
    BUILD = 102
    GAME = 103
    DE_UNKNOWN_104 = 104
    WALL = 105
    DELETE = 106
    ATTACK_GROUND = 107
    TRIBUTE = 108
    DE_UNKNOWN_109 = 109
    REPAIR = 110
    UNGARRISON = 111
    MULTIQUEUE = 112
    GATE = 114
    FLARE = 115
    SPECIAL = 117
    QUEUE = 119
    GATHER_POINT = 120
    SELL = 122
    BUY = 123
    DROP_RELIC = 126
    TOWN_BELL = 127
    BACK_TO_WORK = 128
    DE_QUEUE = 129
    DE_UNKNOWN_130 = 130
    DE_UNKNOWN_131 = 131
    DE_UNKNOWN_134 = 134
    DE_UNKNOWN_135 = 135
    DE_UNKNOWN_136 = 136
    DE_UNKNOWN_137 = 137
    DE_UNKNOWN_138 = 138
    DE_107_B = 140
    DE_TRIBUTE = 196
    POSTGAME = 255


class Postgame(Enum):
    """Postgame types."""
    WORLD_TIME = 1
    LEADERBOARDS = 2

class Age(Enum):
    """Age types."""
    DARK_AGE = 1
    FEUDAL_AGE = 2
    CASTLE_AGE = 3
    IMPERIAL_AGE = 4



"""Embedded Structures.

Various structure can be embedded in the body, without an operation header.

Saved Chapters:

A saved chapter is a header structure embedded in the body.
There is no command identifier, so the command type is actually
the first field of the header - length/offset. Applying the `subheader`
struct at this point will parse the embedded header.


This section is a work in progress.
"""

from construct import (Array, Computed, Embedded, GreedyBytes, If, Int16ul,
                       Int32ul, Padding, Peek, String, Struct, Switch)

from mgz import subheader

# pylint: disable=invalid-name

# Embedded chat message
chat = Struct(
    "subtype"/Computed("chat"),
    "data"/Struct(
        "length"/Computed(lambda ctx: ctx._._._.op),
        "text"/String(lambda ctx: ctx._._._.op, padchar=b'\x00',
                      trimdir='right', encoding='latin1'),
    )
)

# Embedded header (aka saved chapter)
header = Struct(
    "subtype"/Computed("savedchapter"),
    "data"/Struct(
        "header_length"/Computed(lambda ctx: ctx._._._.op - ctx._._._.start),
        Embedded(subheader)
    )
)

# Unknown embedded structure - looks like a partial action?
other = Struct(
    "subtype"/Computed("unknown"),
    "data"/Struct(
        Padding(4),
        "num_ints"/Int32ul,
        If(lambda ctx: ctx.num_ints < 0xff, Array(
            lambda ctx: ctx.num_ints, Int32ul
        )),
        Padding(12)
    )
)

# Anything we don't recognize - just consume the remainder
default = Struct(
    "subtype"/Computed("default"),
    GreedyBytes
)


# Embedded structures identified by first byte (for now)
embedded = "embedded"/Struct(
    "marker"/Peek(Int16ul),
    Embedded("data"/Switch(lambda ctx: ctx.marker, {
        0: header,
        9024: chat,
        65535: other
    }, default=default))
)



"""Action parsing for DE >= 71094."""
import io
import struct

from mgz.util import unpack, as_hex
from mgz.fast.enums import Action


def parse_action_71094(action_type, player_id, raw):
    data = io.BytesIO(raw)
    payload = {}
    if action_type is Action.RESIGN:
        unpack('<b', data)
    if action_type is Action.RESEARCH:
        object_id, selected, technology_id = unpack('<Ihh5x', data)
        selected_building_ids = unpack(f'<{selected}I', data, shorten=False)
        payload = dict(technology_id=technology_id, object_ids=[object_id])
    if action_type is Action.GAME:
        command_id = unpack('<h', data)
        payload = dict(command_id=command_id)
        if command_id == 0:
            source_player, target_player, mode_float, mode = unpack('<2xhhfb', data)
            payload.update(dict(target_player_id=target_player, diplomacy_mode=mode))
        elif command_id == 1:
            payload['speed'] = unpack('<6xf', data)
        elif command_id in [13, 14, 17, 18]:
            payload['number'] = unpack('<4xh', data)
    if action_type is Action.DE_QUEUE:
        selected, building_type, unit_id, amount, *object_ids = unpack('<h4xhhh4x', data)
        object_ids = list(unpack(f'<{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, amount=amount, unit_id=unit_id)
    if action_type is Action.MOVE:
        x, y, selected = unpack('<4x2fh', data)
        object_ids = []
        data.read(6)
        if selected > 0:
            object_ids = list(unpack(f'<{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, x=x, y=y)
    if action_type is Action.ORDER:
        target_id, x, y, selected = unpack('<I2fh', data)
        object_ids = []
        data.read(6)
        if selected > 0:
            object_ids = list(unpack(f'<{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, target_id=target_id, x=x, y=y)
    if action_type is Action.BUILD:
        selected, x, y, building_id, unk2, unk3, unk4 = unpack('<h2xffI8xhbb', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(building_id=building_id, object_ids=object_ids, x=x, y=y)
    if action_type is Action.GATHER_POINT:
        selected, x, y, target_id, target_type = unpack('<h2xffiix', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(target_id=target_id, target_type=target_type, x=x, y=y, object_ids=object_ids)
    if action_type is Action.DE_MULTI_GATHERPOINT:
        target_id, x, y = unpack('<iff', data) # This is a best guess. There is other unknown data in the payload.
        payload = dict(target_id=target_id, x=x, y=y)
    if action_type is Action.STANCE:
        selected, stance_id = unpack('<II', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(stance_id=stance_id, object_ids=object_ids)
    if action_type is Action.SPECIAL:
        selected, target_id, x, y, slot_id, order_id = unpack('<Iiff4xh2xh2x', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(order_id=order_id, slot_id=slot_id, target_id=target_id, x=x, y=y, object_ids=object_ids)
    if action_type is Action.FORMATION:
        selected, formation_id = unpack('<II', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(formation_id=formation_id, object_ids=object_ids)
    if action_type in [Action.BUY, Action.SELL]:
        resource_id, amount, object_id = unpack('<hhI', data)
        payload = dict(resource_id=resource_id, amount=amount, object_ids=[object_id])
    if action_type is Action.DE_TRANSFORM:
        # autoscout enable?
        object_id, y = unpack('<II', data)
        payload = dict(object_ids=[object_id])
    if action_type is Action.AI_ORDER:
        # used for autoscout moves
        # 01 00 00 00 75 06 00 00 ff ff ff ff 21 03 00 00 00 00 30 42 00 00 a0 42 00 00 00 00 00 00 80 3f 64 ff 01 00
        a, object_id, c, x, y = unpack('<II4xIff', data)
        payload = dict(object_ids=[object_id], x=x, y=y)
    if action_type in [Action.BACK_TO_WORK, Action.DELETE]:
        object_id = unpack('<I', data)
        payload = dict(object_ids=[object_id])
    if action_type is Action.WALL:
        selected, x1, y1, x2, y2, building_id = unpack('<IHHHHI', data)
        data.read(8)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, x=x1, y=y1, x_end=x2, y_end=y2, building_id=building_id)
    if action_type in [Action.PATROL, Action.DE_ATTACK_MOVE]:
        selected, x, y = unpack('<I4xf36xf36x', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, x=x, y=y)
    if action_type is Action.UNGARRISON:
        selected, x, y, target_id, unk = unpack('<IffiI', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, x=x, y=y, target_id=target_id)
    if action_type is Action.FLARE:
        x, y, num = unpack('<4xffb', data)
        targets = list(unpack(f'<{num}b', data, shorten=False))
        payload = dict(x=x, y=y, targets=targets)
    if action_type is Action.TOWN_BELL:
        building_id, mode = unpack('<Ib',data)
        payload = dict(building_id=building_id, mode=mode)
    if action_type is Action.STOP:
        selected = unpack('<I', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids)
    if action_type in [Action.FOLLOW, Action.GUARD]:
        selected, target_id = unpack('<II', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, target_id=target_id)
    if action_type is Action.ATTACK_GROUND:
        selected, x, y = unpack('<Iff', data)
        data.read(4)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, x=x, y=y)
    if action_type is Action.REPAIR:
        selected, target_id = unpack('<II', data)
        data.read(4)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids, target_id=target_id)
    if action_type is Action.DE_TRIBUTE:
        wood, food, gold, stone = unpack('<ffff', data)
        data.read(16) # cost[4]
        data.read(8) # attribute id[4]
        target_id = data.read(1)
        payload = dict(target_player_id=target_id, food=food, wood=wood, stone=stone, gold=gold)
    if action_type in [Action.GATE, Action.DROP_RELIC]:
        object_id = unpack('<I', data)
        payload = dict(object_ids=[object_id])
    if action_type in [Action.DE_AUTOSCOUT, Action.RATHA_ABILITY]:
        selected = unpack('<I', data)
        object_ids = list(unpack(f'{selected}I', data, shorten=False))
        payload = dict(object_ids=object_ids)
    if action_type is Action.MAKE:
        building_id, unit_id = unpack('<H6xh', data)
        payload = dict(building_id=building_id, unit_id=unit_id)
    return dict(player_id=player_id, **payload)



"""MGZ Summary."""

import asyncio
import collections
import hashlib
import logging
import os
import io
import json
import struct
import tempfile
import time
import uuid
import zlib

import construct

import mgz
import mgz.body
from mgz import fast
from mgz.util import Version

from mgz.const import VALID_BUILDINGS
from mgz.common.map import get_map_data
from mgz.summary.settings import get_settings_data
from mgz.summary.dataset import get_dataset_data
from mgz.summary.teams import get_teams_data
from mgz.summary.players import get_players_data, enrich_de_player_data
from mgz.common.diplomacy import get_diplomacy_data
from mgz.common.chat import get_lobby_chat, parse_chat, Chat
from mgz.summary.objects import get_objects_data


LOGGER = logging.getLogger(__name__)
CHECKSUMS = 4
MAX_SYNCS = 2000
AI_ACTIONS = [fast.Action.AI_ORDER]


class FullSummary: # pylint: disable=too-many-public-methods
    """MGZ summary.

    Access match summary data.
    """

    def __init__(self, handle):
        """Initialize."""
        self.size = len(handle.read())
        handle.seek(0)
        self._handle = handle
        self._cache = {
            'dataset': None,
            'teams': None,
            'resigned': set(),
            'cheaters': set(),
            'encoding': None,
            'language': None,
            'ratings': {},
            'postgame': None,
            'from_voobly': False,
            'platform_id': None,
            'platform_match_id': None,
            'rated': None,
            'ladder': None,
            'hash': None,
            'map': None,
            'lobby_name': None,
            'duration': None,
            'extraction': None
        }
        self._eapm = collections.Counter()

        try:
            start = time.time()
            self._header = mgz.header.parse_stream(self._handle)
            LOGGER.info("parsed header in %.2f seconds", time.time() - start)
            self._chats = get_lobby_chat(
                self._header, self.get_encoding(),
                self.get_diplomacy().get('type'), self.get_players()
            )
            body_pos = self._handle.tell()
            self._cache['file_hash'] = hashlib.sha1(self._handle.read()).hexdigest()
            self._handle.seek(body_pos)
            self._process_body()
            self.body_pos = body_pos
        except (construct.core.ConstructError, zlib.error, ValueError) as e:
            raise RuntimeError("invalid mgz file: {}".format(e))

    def _process_body(self): # pylint: disable=too-many-locals, too-many-statements, too-many-branches
        """Process rec body."""
        start_time = time.time()
        ratings = {}
        checksums = []
        ladder = None
        voobly = False
        rated = None
        i = 0
        duration = self._header.initial.restore_time
        fast.meta(self._handle)
        self._actions = []
        while True:
            try:
                operation, payload = fast.operation(self._handle)
                if operation == fast.Operation.SYNC:
                    i += 1
                    duration += payload[0]
                    if payload[1] and len(checksums) < CHECKSUMS:
                        checksums.append(payload[1].to_bytes(8, 'big', signed=True))
                elif operation == fast.Operation.ACTION:
                    if 'player_id' in payload[1] and payload[0] not in AI_ACTIONS:
                        self._eapm[payload[1]['player_id']] += 1
                    self._actions.append((duration, *payload))
                    if payload[0] == fast.Action.POSTGAME:
                        self._cache['postgame'] = mgz.body.actions.postgame.parse(payload[1]['bytes'])
                    elif payload[0] == fast.Action.RESIGN:
                        self._cache['resigned'].add(payload[1]['player_id'])
                    elif payload[0] == fast.Action.TRIBUTE and 'player_id_to' in payload[1] and payload[1]['player_id_to'] == 0:
                        self._cache['cheaters'].add(payload[1]['player_id'])
                    elif payload[0] == fast.Action.TRIBUTE and payload[1]['player_id'] == 0:
                        self._cache['cheaters'].add(payload[1]['player_id_to'])
                    elif payload[0] == fast.Action.CREATE:
                        self._cache['cheaters'].add(payload[1]['player_id'])
                    elif payload[0] == fast.Action.BUILD and payload[1]['building_id'] not in VALID_BUILDINGS:
                        self._cache['cheaters'].add(payload[1]['player_id'])
                    elif payload[0] == fast.Action.GAME and payload[1]['command_id'] in [2, 4, 6]:
                        self._cache['cheaters'].add(payload[1]['player_id'])
                elif operation == fast.Operation.CHAT:
                    text = payload
                    if text is None:
                        continue
                    try:
                        parsed = parse_chat(
                            text, self.get_encoding(), duration, self.get_players(), self.get_diplomacy().get('type')
                        )
                        self._chats.append(parsed)
                        if parsed['type'] == Chat.RATING:
                            ratings[parsed['player']] = parsed['rating']
                        elif parsed['type'] == Chat.LADDER:
                            ladder = parsed['ladder']
                        elif parsed['type'] == Chat.VOOBLY:
                            voobly = True
                    except UnicodeDecodeError:
                        pass
            except EOFError:
                break
        self._cache['duration'] = duration
        if voobly:
            rated = len(ratings) > 0 and set(ratings.values()) != {1600}
        if self._header.version == Version.DE:
            self._cache['hash'] = hashlib.sha1(self._header.de.guid)
        elif self._header.version == Version.HD and self._header.save_version >= 12.49:
            self._cache['hash'] = hashlib.sha1(self._header.hd.guid)
        else:
            self._cache['hash'] = hashlib.sha1(b''.join(checksums)) \
                if len(checksums) == CHECKSUMS else None
        if self._header.de:
            rated = self._header.de.ranked
        self._cache['from_voobly'] = voobly
        if voobly:
            self._cache['platform_id'] = 'voobly'
        if self._header.version == Version.DE and self._header.de.multiplayer:
            self._cache['platform_id'] = 'de'
        if self._header.version == Version.HD and self._header.hd.multiplayer:
            self._cache['platform_id'] = 'hd'
        self._cache['ladder'] = ladder
        self._cache['rated'] = rated
        self._cache['ratings'] = ratings if rated else {}
        LOGGER.info("parsed body in %.2f seconds", time.time() - start_time)

    def get_chat(self):
        """Get chat messages."""
        return self._chats

    def get_postgame(self):
        """Get postgame structure."""
        return self._cache['postgame']

    def has_achievements(self):
        """If match has achievements available."""
        return self._cache['postgame'] is not None or self._cache['extraction'] is not None

    def get_header(self):
        """Get header."""
        return self._header

    def get_start_time(self):
        """Get match start time delta."""
        return self._header.initial.restore_time

    def get_duration(self):
        """Get game duration."""
        return self._cache['duration']

    def get_restored(self):
        """Check for restored game."""
        return self._header.initial.restore_time > 0, self._header.initial.restore_time

    def get_version(self):
        """Get game version."""
        build = None
        if self._header.de:
            build = self._header.de.build
        return self._header.version, self._header.game_version, self._header.save_version, self._header.log_version, build

    def get_owner(self):
        """Get rec owner (POV)."""
        return self._header.replay.rec_player

    def get_teams(self):
        """Get teams."""
        if not self._cache['teams']:
            self._cache['teams'] = get_teams_data(self._header)
        return self._cache['teams']

    def get_diplomacy(self):
        """Compute diplomacy."""
        return get_diplomacy_data(self.get_header(), self.get_teams())

    def get_profile_ids(self):
        """Get map of player color to profile IDs (DE/HD only)."""
        if self._header.version == Version.DE:
            key = 'de'
            field = 'profile_id'
        elif self._header.version == Version.HD and self._header.save_version >= 12.49:
            key = 'hd'
            field = 'steam_id'
        else:
            return {}
        return {
            p.player_number: p[field]
            for p in self._header[key].players
            if p.player_number >= 0 and p[field] > 0
        }

    def get_players(self):
        """Get players."""
        data = get_players_data(
            self.get_header(),
            self.get_postgame(),
            self.get_teams(),
            self._cache['resigned'],
            self._cache['cheaters'],
            self.get_profile_ids(),
            self.get_ratings(),
            self.get_encoding(),
            {id_:int(round(v/((self.get_duration()/1000)/60))) for id_, v in self._eapm.items() if self.get_duration() and v}
        )
        if self._cache['extraction']:
            enrich_de_player_data(data, self._cache['extraction'])
        return data

    def get_objects(self):
        """Get objects."""
        return get_objects_data(self._header)

    def get_ratings(self):
        """Get player ratings."""
        if not self._cache['ratings']:
            self.get_platform()
        return self._cache['ratings']

    def get_platform(self):
        """Get platform data."""
        lobby_name = None
        guid = None
        if self._header.version == Version.DE:
            lobby_name = self._header.de.lobby_name.value.decode(self.get_encoding()).strip()
            guid = str(uuid.UUID(bytes=self._header.de.guid))
        elif self._header.version == Version.HD and self._header.save_version >= 12.49:
            lobby_name = self._header.hd.lobby_name.value.decode(self.get_encoding()).strip()
            guid = str(uuid.UUID(bytes=self._header.hd.guid))
            rating_key = "hd_{}_rating".format(self._header.lobby.game_type.lower())
            for player in self._header.hd.players:
                if player.player_number < 1:
                    continue
                if rating_key not in player:
                    continue
                self._cache['ratings'][player.name.value.decode(self.get_encoding())] = player[rating_key]
            self._cache['rated'] = self._header.hd.is_ranked
        return {
            'platform_id': self._cache['platform_id'],
            'platform_match_id': guid,
            'ladder': self._cache['ladder'],
            'rated': self._cache['rated'],
            'ratings': self._cache['ratings'],
            'lobby_name': lobby_name,
            'allow_specs': self._header.de.allow_specs if self._header.de else None,
            'spec_delay': self._header.de.spec_delay if self._header.de else None,
            'private': self._header.de.lobby_visibility == 2 if self._header.de else None,
        }

    def get_settings(self):
        """Get settings."""
        return get_settings_data(self.get_postgame(), self._header)

    def get_file_hash(self):
        """Get file hash."""
        return self._cache['file_hash']

    def get_hash(self):
        """Get cached hash."""
        return self._cache['hash']

    def get_encoding(self):
        """Get text encoding."""
        if not self._cache['encoding']:
            self.get_map()
        return self._cache['encoding']

    def get_language(self):
        """Get language."""
        if not self._cache['language']:
            self.get_map()
        return self._cache['language']

    def get_device(self):
        """Get device."""
        return None

    def get_map_id(self):
        """Get map ID."""
        if self._header.hd:
            return self._header.hd.selected_map_id
        if self._header.de:
            return self._header.de.resolved_map_id
        return self._header.scenario.game_settings.map_id

    def get_map(self):
        """Get map."""
        tiles = [(tile.terrain_type, tile.elevation) for tile in self._header.map_info.tile]
        if not self._cache['map']:
            self._cache['map'], self._cache['encoding'], self._cache['language'] = get_map_data(
                self.get_map_id(),
                self._header.scenario.messages.instructions,
                self._header.map_info.size_x,
                self._header.version,
                self.get_dataset()['id'],
                self.reference,
                tiles,
                de_seed=self._header.lobby.de.map_seed if self._header.lobby.de else None,
                de_strings=self._header.de.rms_strings.strings if self._header.de else []
            )
        return self._cache['map']

    def get_dataset(self):
        """Get dataset."""
        if not self._cache['dataset']:
            self._cache['dataset'] = get_dataset_data(self._header)
        self.reference = self._cache['dataset'][1]
        return self._cache['dataset'][0]

    def get_completed(self):
        """Determine if the game was completed.

        If there's a postgame, it will indicate completion.
        If there is no postgame, guess based on resignation.
        """
        postgame = self.get_postgame()
        if postgame:
            return postgame.complete
        return bool(self._cache['resigned'])

    def get_mirror(self):
        """Determine mirror match."""
        mirror = False
        if self.get_diplomacy()['type'] == '1v1':
            civs = set()
            for data in self.get_players():
                civs.add(data['civilization'])
            mirror = (len(civs) == 1)
        return mirror

    def get_played(self):
        if self._header.de:
            return self._header.de.timestamp



from construct import (
    Struct, Int32ul, Float32l, Array, Padding, Flag, If,
    Byte, Int16ul, Bytes, Int32sl, Peek, Const, RepeatUntil,
    Int64ul, Computed, Embedded, IfThenElse
)

from mgz.enums import VictoryEnum, ResourceLevelEnum, AgeEnum, PlayerTypeEnum, DifficultyEnum
from mgz.util import find_save_version

separator = Const(b"\xa3_\x02\x00")

hd_string = Struct(
    "length"/Int16ul,
    Const(b"\x60\x0A"),
    "value"/Bytes(lambda ctx: ctx.length)
)

test_57 = "test_57"/Struct(
    "check"/Int32ul,
    Padding(4),
    If(lambda ctx: ctx._._.version >= 1006, Bytes(1)),
    Padding(15),
    hd_string,
    Padding(1),
    If(lambda ctx: ctx._._.version >= 1005, hd_string),
    hd_string,
    Padding(16),
    "test"/Int32ul,
    "is_57"/Computed(lambda ctx: ctx.check == ctx.test)
)

player = Struct(
    "dlc_id"/Int32ul,
    "color_id"/Int32ul,
    "unk1_1006"/If(lambda ctx: ctx._._.version >= 1006, Bytes(1)),
    "unk"/Bytes(2),
    "dat_crc"/Bytes(4),
    "mp_game_version"/Byte,
    "team_index"/Int32ul,
    "civ_id"/Int32ul,
    "ai_type"/hd_string,
    "ai_civ_name_index"/Byte,
    "ai_name"/If(lambda ctx: ctx._._.version >= 1005, hd_string),
    "name"/hd_string,
    "type"/PlayerTypeEnum(Int32ul),
    "steam_id"/Int64ul,
    "player_number"/Int32sl,
    Embedded(If(lambda ctx: ctx._._.version >= 1006 and not ctx._.test_57.is_57, Struct(
        "hd_rm_rating"/Int32ul,
        "hd_dm_rating"/Int32ul,
    )))
)

hd = "hd"/Struct(
    "version"/Float32l,
    "interval_version"/Int32ul,
    "game_options_version"/Int32ul,
    "dlc_count"/Int32ul,
    "dlc_ids"/Array(lambda ctx: ctx.dlc_count, Int32ul),
    "dataset_ref"/Int32ul,
    Peek("difficulty_id"/Int32ul),
    DifficultyEnum("difficulty"/Int32ul),
    "selected_map_id"/Int32ul,
    "resolved_map_id"/Int32ul,
    "reveal_map"/Int32ul,
    Peek("victory_type_id"/Int32ul),
    VictoryEnum("victory_type"/Int32ul),
    Peek("starting_resources_id"/Int32ul),
    ResourceLevelEnum("starting_resources"/Int32ul),
    "starting_age_id"/Int32ul,
    "starting_age"/AgeEnum(Computed(lambda ctx: ctx.starting_age_id)),
    "ending_age_id"/Int32ul,
    "ending_age"/AgeEnum(Computed(lambda ctx: ctx.ending_age_id)),
    "game_type"/If(lambda ctx: ctx.version >= 1006, Int32ul),
    separator,
    "ver1000"/If(lambda ctx: ctx.version == 1000, Struct(
        "map_name"/hd_string,
        "unk"/hd_string
    )),
    separator,
    "speed"/Float32l,
    "treaty_length"/Int32ul,
    "population_limit"/Int32ul,
    "num_players"/Int32ul,
    "unused_player_color"/Int32ul,
    "victory_amount"/Int32ul,
    separator,
    "trade_enabled"/Flag,
    "team_bonus_disabled"/Flag,
    "random_positions"/Flag,
    "all_techs"/Flag,
    "num_starting_units"/Byte,
    "lock_teams"/Flag,
    "lock_speed"/Flag,
    "multiplayer"/Flag,
    "cheats"/Flag,
    "record_game"/Flag,
    "animals_enabled"/Flag,
    "predators_enabled"/Flag,
    "turbo_enabled"/Flag,
    "shared_exploration"/Flag,
    "team_positions"/Flag,
    "unk"/Bytes(1),
    Embedded(IfThenElse(lambda ctx: ctx.version == 1000,
        Struct(
            Bytes(40*3),
            separator,
            Bytes(40),
            "strings"/Array(8, hd_string),
            Bytes(16),
            separator,
            Bytes(10),
        ),
        Struct(
            Peek(test_57),
            "players"/Array(8, player),
            "fog_of_war"/Flag,
            "cheat_notifications"/Flag,
            "colored_chat"/Flag,
            Bytes(9),
            separator,
            "is_ranked"/Flag,
            "allow_specs"/Flag,
            "lobby_visibility"/Int32ul,
            "custom_random_map_file_crc"/Int32ul,
            "custom_scenario_or_campaign_file"/hd_string,
            Bytes(8),
            "custom_random_map_file"/hd_string,
            Bytes(8),
            "custom_random_map_scenarion_file"/hd_string,
            Bytes(8),
            "guid"/Bytes(16),
            "lobby_name"/hd_string,
            "modded_dataset"/hd_string,
            "modded_dataset_workshop_id"/Bytes(4),
            If(lambda ctx: ctx._.version >= 1005,
                Struct(
                    Bytes(4),
                    hd_string,
                    Bytes(4)
                )
            )
        )
    ))
)
